"""One common way to define how similar two datasets are is to compute some "similarity"
between them. This notion of "similarity" requires computing some features of a dataset
(**metafeatures**) first, such that we can numerically compute some distance function.

Let's see how we can quickly compute the distance between some datasets with
[`dataset_distance()`][amltk.metalearning.dataset_distance]!

```python exec="true" source="material-block" result="python" title="Dataset Distances P.1" session='dd'
import pandas as pd
import openml

from amltk.metalearning import compute_metafeatures

def get_dataset(dataset_id: int) -> tuple[pd.DataFrame, pd.Series]:
    dataset = openml.datasets.get_dataset(
        dataset_id,
        download_data=True,
        download_features_meta_data=False,
        download_qualities=False,
    )
    X, y, _, _ = dataset.get_data(
        dataset_format="dataframe",
        target=dataset.default_target_attribute,
    )
    return X, y

d31 = get_dataset(31)
d3 = get_dataset(3)
d4 = get_dataset(4)

metafeatures_dict = {
    "dataset_31": compute_metafeatures(*d31),
    "dataset_3": compute_metafeatures(*d3),
    "dataset_4": compute_metafeatures(*d4),
}

metafeatures = pd.DataFrame(metafeatures_dict)
print(metafeatures)
```

Now we want to know which one of `#!python "dataset_3"` or `#!python "dataset_4"` is
more _similar_ to `#!python "dataset_31"`.

```python exec="true" source="material-block" result="python" title="Dataset Distances P.2" session='dd'
from amltk.metalearning import dataset_distance

target = metafeatures_dict.pop("dataset_31")
others = metafeatures_dict

distances = dataset_distance(target, others, distance_metric="l2")
print(distances)
```

Seems like `#!python "dataset_3"` is some notion of closer to `#!python "dataset_31"`
than `#!python "dataset_4"`. However the scale of the metafeatures are not exactly all close.
For example, many lie between `#!python (0, 1)` but some like `instance_count` can completely
dominate the show.

Lets repeat the computation but specify that we should apply a `#!python "minmax"` scaling
across the rows.

```python exec="true" source="material-block" result="python" title="Dataset Distances P.3" session='dd' hl_lines="5"
distances = dataset_distance(
    target,
    others,
    distance_metric="l2",
    scaler="minmax"
)
print(distances)
```

Now `#!python "dataset_3"` is considered more similar but the difference between the two is a lot less
dramatic. In general, applying some scaling to values of different scales is required for metalearning.

You can also use an [sklearn.preprocessing.MinMaxScaler][] or anything other scaler from scikit-learn
for that matter.

```python exec="true" source="material-block" result="python" title="Dataset Distances P.3" session='dd' hl_lines="7"
from sklearn.preprocessing import MinMaxScaler

distances = dataset_distance(
    target,
    others,
    distance_metric="l2",
    scaler=MinMaxScaler()
)
print(distances)
```
"""  # noqa: E501
from __future__ import annotations

import warnings
from collections.abc import Callable, Mapping
from typing import TYPE_CHECKING, Literal, TypeVar

import pandas as pd

from amltk._functional import funcname
from amltk.distances import (
    DistanceMetric,
    NamedDistance,
    NearestNeighborsDistance,
    distance_metrics,
)
from amltk.types import safe_isinstance

if TYPE_CHECKING:
    from sklearn.base import TransformerMixin


D = TypeVar("D")

DatasetDistanceMetric = Callable[[pd.DataFrame, pd.Series], pd.Series]


def _metric_for_frame(
    _metric: DistanceMetric,
) -> Callable[[pd.DataFrame, pd.Series], pd.Series]:
    def _new_metric(x: pd.DataFrame, y: pd.Series) -> pd.Series:
        _x = x.apply(lambda col: _metric(col.to_numpy(), y.to_numpy()))
        assert isinstance(_x, pd.Series)
        return _x

    return _new_metric


def dataset_distance(  # noqa: C901, PLR0912
    target: pd.Series,
    dataset_metafeatures: Mapping[str, pd.Series],
    *,
    distance_metric: (DistanceMetric | NearestNeighborsDistance | NamedDistance) = "l2",
    scaler: TransformerMixin
    | Callable[[pd.DataFrame], pd.DataFrame]
    | Literal["minmax"]
    | None = None,
    closest_n: int | None = None,
) -> pd.Series:
    """Calculates the distance between a target dataset and a set of datasets.

    This uses the metafeatures of the datasets to calculate the distance.

    Args:
        target: The target dataset's metafeatures.
        dataset_metafeatures: A dictionary of dataset names to their metafeatures.
        distance_metric: The method to use to calculate the distance.
            Takes in the target dataset's metafeatures and a dataset's metafeatures
            Should return the distance between the two.
        scaler: A scaler to use to scale the metafeatures.
        closest_n: The number of closest datasets to return. If None, all datasets
            are returned.

    Returns:
        Series with the index being the dataset name and the values being the distance.
    """
    outname: str
    if isinstance(distance_metric, str):
        outname = distance_metric
    else:
        outname = funcname(distance_metric)

    if target.name is None:
        target = target.copy()
        target.name = "target-dataset"

    _method = (
        distance_metrics[distance_metric]
        if isinstance(distance_metric, str)
        else distance_metric
    )

    if not isinstance(_method, NearestNeighborsDistance):
        _method = _metric_for_frame(_method)

    metafeatures = {
        name: ds_metafeatures.rename(name)
        for name, ds_metafeatures in dataset_metafeatures.items()
    }

    # Index is dataset name with columns being the values
    #      | mf1 | mf2
    # d1
    # d2
    # d3
    combined = pd.concat([target, *metafeatures.values()], axis=1).T

    if scaler is None:
        pass
    elif scaler == "minmax":
        min_maxs = combined.agg(["min", "max"], axis=0).T

        mins = min_maxs["min"]
        maxs = min_maxs["max"]
        normalizer = maxs - mins
        normalizer[normalizer == 0] = 1
        mins[normalizer == 0] = 0

        norm = lambda col: (col - mins) / normalizer
        combined = combined.apply(norm, axis=1)
    elif safe_isinstance(scaler, "TransformerMixin"):
        combined = scaler.set_output(transform="pandas").fit_transform(  # type: ignore
            combined,
        )
    elif callable(scaler):
        combined = scaler(combined)
    else:
        raise ValueError(f"Unsure how to handle {scaler=}")

    # We now transpose the dataframe so that the index is the metafeature name
    # while the columns are the dataset names
    #   x   | d1 | d2 | d3          y | dy
    #  mf1                      mf1
    #  mf2                      mf2
    x = combined.T.drop(columns=target.name)
    y = combined.loc[target.name]

    # Should return a series with index being dataset names and values being the
    #     | distance
    # d1
    # d2
    dataset_distances = _method(x, y)

    if not isinstance(dataset_distances, pd.Series):
        dataset_distances = pd.Series(
            dataset_distances,
            dtype=float,
            index=list(dataset_metafeatures.keys()),
            name=outname,
        )
    else:
        dataset_distances = dataset_distances.astype(float).rename(outname)

    dataset_distances = dataset_distances.sort_values()

    if closest_n is not None:
        if closest_n > len(dataset_distances):
            warnings.warn(
                f"Cannot get {closest_n} closest datasets when there are"
                f" only {len(dataset_distances)} datasets. Returning all.",
                UserWarning,
                stacklevel=2,
            )

        dataset_distances = dataset_distances.iloc[:closest_n]

    return dataset_distances


if __name__ == "__main__":
    target = pd.Series([1, 1], name="target", index=["mf1", "mf2"])
    dataset_metafeatures = {
        "d1": pd.Series([1, 1], name="d1", index=["mf1", "mf2"]),
        "d2": pd.Series([-1, -1], name="d2", index=["mf1", "mf2"]),
        "d3": pd.Series([1, -1], name="d2", index=["mf1", "mf2"]),
    }
    distances = dataset_distance(
        target,
        dataset_metafeatures,
        distance_metric=NearestNeighborsDistance(algorithm="brute", metric="l1"),
    )

    distances = dataset_distance(
        target,
        dataset_metafeatures,
        distance_metric="cosine",
    )



from __future__ import annotations

from amltk.metalearning.dataset_distances import dataset_distance
from amltk.metalearning.metafeatures import (
    ClassCounts,
    ClassImbalance,
    ClassImbalanceRatios,
    DatasetStatistic,
    ImbalancePerCategory,
    KurtosisMax,
    KurtosisMean,
    KurtosisMin,
    KurtosisPerNumericalColumn,
    KurtosisStd,
    LogInstanceCount,
    LogNumberOfFeatures,
    MajorityClassImbalance,
    MeanCategoricalImbalance,
    MetaFeature,
    MinorityClassImbalance,
    NumberOfCategoricalFeatures,
    NumberOfClasses,
    NumberOfFeatures,
    NumberOfNumericFeatures,
    PercentageMissingValues,
    PercentageOfCategoricalColumnsWithMissingValues,
    PercentageOfCategoricalValuesWithMissingValues,
    PercentageOfFeaturesWithMissingValues,
    PercentageOfInstancesWithMissingValues,
    PercentageOfNumericColumnsWithMissingValues,
    PercentageOfNumericValuesWithMissingValues,
    RatioCategoricalFeatures,
    RatioFeaturesToInstances,
    RatioNumericalFeatures,
    SkewnessMax,
    SkewnessMean,
    SkewnessMin,
    SkewnessPerNumericalColumn,
    SkewnessStd,
    StdCategoricalImbalance,
    compute_metafeatures,
    metafeature_descriptions,
)
from amltk.metalearning.portfolio import portfolio_selection


def get_metafeatures() -> list[type[DatasetStatistic]]:
    """Returns a list of all available metafeatures."""
    return [
        ClassCounts,
        ClassImbalance,
        ClassImbalanceRatios,
        ImbalancePerCategory,
        KurtosisMax,
        KurtosisMean,
        KurtosisMin,
        KurtosisPerNumericalColumn,
        KurtosisStd,
        LogInstanceCount,
        LogNumberOfFeatures,
        MajorityClassImbalance,
        MeanCategoricalImbalance,
        MinorityClassImbalance,
        NumberOfCategoricalFeatures,
        NumberOfClasses,
        NumberOfFeatures,
        NumberOfNumericFeatures,
        PercentageMissingValues,
        PercentageOfCategoricalColumnsWithMissingValues,
        PercentageOfCategoricalValuesWithMissingValues,
        PercentageOfFeaturesWithMissingValues,
        PercentageOfInstancesWithMissingValues,
        PercentageOfNumericColumnsWithMissingValues,
        PercentageOfNumericValuesWithMissingValues,
        RatioFeaturesToInstances,
        RatioNumericalFeatures,
        RatioCategoricalFeatures,
        SkewnessMax,
        SkewnessMean,
        SkewnessMin,
        SkewnessPerNumericalColumn,
        SkewnessStd,
        StdCategoricalImbalance,
    ]


__all__ = [
    "ClassCounts",
    "ClassImbalance",
    "DatasetStatistic",
    "ImbalancePerCategory",
    "KurtosisMax",
    "KurtosisMean",
    "KurtosisMin",
    "KurtosisPerNumericalColumn",
    "KurtosisStd",
    "LogInstanceCount",
    "LogNumberOfFeatures",
    "MajorityClassImbalance",
    "MeanCategoricalImbalance",
    "MetaFeature",
    "MinorityClassImbalance",
    "NumberOfCategoricalFeatures",
    "NumberOfClasses",
    "NumberOfFeatures",
    "NumberOfNumericFeatures",
    "PercentageMissingValues",
    "PercentageOfCategoricalColumnsWithMissingValues",
    "PercentageOfCategoricalValuesWithMissingValues",
    "PercentageOfFeaturesWithMissingValues",
    "PercentageOfInstancesWithMissingValues",
    "PercentageOfNumericColumnsWithMissingValues",
    "PercentageOfNumericValuesWithMissingValues",
    "RatioCategoricalFeatures",
    "RatioFeaturesToInstances",
    "RatioNumericalFeatures",
    "SkewnessMax",
    "SkewnessMean",
    "SkewnessMin",
    "SkewnessPerNumericalColumn",
    "SkewnessStd",
    "StdCategoricalImbalance",
    "compute_metafeatures",
    "dataset_distance",
    "get_metafeatures",
    "metafeature_descriptions",
    "portfolio_selection",
]



"""One of the primary ways to respond to `@events` emitted
with by a [`Task`][amltk.scheduling.Task]
the [`Scheduler`][amltk.scheduling.Scheduler]
is through use of a **callback**.

The reason for this is to enable an easier time for API's to utilize
multiprocessing and remote compute from the `Scheduler`, without having
to burden users with knowing the details of how to use multiprocessing.

A callback subscribes to some event using a decorator but can also be done in
a functional style if preferred. The below example is based on the
event [`@scheduler.on_start`][amltk.scheduling.Scheduler.on_start] but
the same applies to all events.

=== "Decorators"

    ```python exec="true" source="material-block" html="true"
    from amltk.scheduling import Scheduler

    scheduler = Scheduler.with_processes(1)

    @scheduler.on_start
    def print_hello() -> None:
        print("hello")

    scheduler.run()
    from amltk._doc import doc_print; doc_print(print, scheduler, fontsize="small")  # markdown-exec: hide
    ```

=== "Functional"

    ```python exec="true" source="material-block" html="true"
    from amltk.scheduling import Scheduler

    scheduler = Scheduler.with_processes(1)

    def print_hello() -> None:
        print("hello")

    scheduler.on_start(print_hello)
    scheduler.run()
    from amltk._doc import doc_print; doc_print(print, scheduler, fontsize="small")  # markdown-exec: hide
    ```

There are a number of ways to customize the behaviour of these callbacks, notably
to control how often they get called and when they get called.

??? tip "Callback customization"


    === "`on('event', repeat=...)`"

        This will cause the callback to be called `repeat` times successively.
        This is most useful in combination with
        [`@scheduler.on_start`][amltk.scheduling.Scheduler.on_start] to launch
        a number of tasks at the start of the scheduler.

        ```python exec="true" source="material-block" html="true" hl_lines="11"
        from amltk import Scheduler

        N_WORKERS = 2

        def f(x: int) -> int:
            return x * 2
        from amltk._doc import make_picklable; make_picklable(f)  # markdown-exec: hide

        scheduler = Scheduler.with_processes(N_WORKERS)
        task = scheduler.task(f)

        @scheduler.on_start(repeat=N_WORKERS)
        def on_start():
            task.submit(1)

        scheduler.run()
        from amltk._doc import doc_print; doc_print(print, scheduler, fontsize="small")  # markdown-exec: hide
        ```

    === "`on('event', max_calls=...)`"

        Limit the number of times a callback can be called, after which, the callback
        will be ignored.

        ```python exec="true" source="material-block" html="True" hl_lines="13"
        from asyncio import Future
        from amltk.scheduling import Scheduler

        scheduler = Scheduler.with_processes(2)

        def expensive_function(x: int) -> int:
            return x ** 2
        from amltk._doc import make_picklable; make_picklable(expensive_function)  # markdown-exec: hide

        @scheduler.on_start
        def submit_calculations() -> None:
            scheduler.submit(expensive_function, 2)

        @scheduler.on_future_result(max_calls=3)
        def print_result(future, result) -> None:
            scheduler.submit(expensive_function, 2)

        scheduler.run()
        from amltk._doc import doc_print; doc_print(print, scheduler, output="html", fontsize="small")  # markdown-exec: hide
        ```

    === "`on('event', when=...)`"

        A callable which takes no arguments and returns a `bool`. The callback
        will only be called when the `when` callable returns `True`.

        Below is a rather contrived example, but it shows how we can use the
        `when` parameter to control when the callback is called.

        ```python exec="true" source="material-block" html="True" hl_lines="8 12"
        import random
        from amltk.scheduling import Scheduler

        LOCALE = random.choice(["English", "German"])

        scheduler = Scheduler.with_processes(1)

        @scheduler.on_start(when=lambda: LOCALE == "English")
        def print_hello() -> None:
            print("hello")

        @scheduler.on_start(when=lambda: LOCALE == "German")
        def print_guten_tag() -> None:
            print("guten tag")

        scheduler.run()
        from amltk._doc import doc_print; doc_print(print, scheduler, output="html", fontsize="small")  # markdown-exec: hide
        ```

    === "`on('event', every=...)`"

        Only call the callback every `every` times the event is emitted. This
        includes the first time it's called.

        ```python exec="true" source="material-block" html="True" hl_lines="6"
        from amltk.scheduling import Scheduler

        scheduler = Scheduler.with_processes(1)

        # Print "hello" only every 2 times the scheduler starts.
        @scheduler.on_start(every=2)
        def print_hello() -> None:
            print("hello")

        # Run the scheduler 5 times
        scheduler.run()
        scheduler.run()
        scheduler.run()
        scheduler.run()
        scheduler.run()
        from amltk._doc import doc_print; doc_print(print, scheduler, output="html", fontsize="small")  # markdown-exec: hide
        ```

### Emitter, Subscribers and Events
This part of the documentation is not necessary to understand or use for AMLTK. People
wishing to build tools upon AMLTK may still find this a useful component to add to their
arsenal.

The core of making this functionality work is the [`Emitter`][amltk.scheduling.events.Emitter].
Its purpose is to have `@events` that can be emitted and subscribed to. Classes like the
[`Scheduler`][amltk.scheduling.Scheduler] and [`Task`][amltk.scheduling.Task] carry
around with them an `Emitter` to enable all of this functionality.

Creating an `Emitter` is rather straight-forward, but we must also create
[`Events`][amltk.scheduling.events.Event] that people can subscribe to.

```python
from amltk.scheduling import Emitter, Event
emitter = Emitter("my-emitter")

event: Event[int] = Event("my-event") # (1)!

@emitter.on(event)
def my_callback(x: int) -> None:
    print(f"Got {x}!")

emitter.emit(event, 42) # (2)!
```

1. The typing `#!python Event[int]` is used to indicate that the event will be emitting
    an integer. This is not necessary, but it is useful for type-checking and
    documentation.
2. The `#!python emitter.emit(event, 42)` is used to emit the event. This will call
    all the callbacks registered for the event, i.e. `#!python my_callback()`.

!!! warning "Independent Events"

    Given a single `Emitter` and a single instance of an `Event`, there is no way to
    have different `@events` for callbacks. There are two options, both used extensively
    in AMLTK.

    The first is to have different `Events` quite naturally, i.e. you distinguish
    between different things that can happen. However, you often want to have different
    objects emit the same `Event` but have different callbacks for each object.

    This makes most sense in the context of a `Task` the `Event` instances are shared as
    class variables in the `Task` class, however a user likely want's to subscribe to
    the `Event` for a specific instance of the `Task`.

    This is where the second option comes in, in which each object carries around its
    own `Emitter` instance. This is how a user can subscribe to the same kind of `Event`
    but individually for each `Task`.


However, to shield users from this and to create named access points for users to
subscribe to, we can use the [`Subscriber`][amltk.scheduling.events.Subscriber] class,
conveniently created by the [`Emitter.subscriber()`][amltk.scheduling.events.Emitter.subscriber]
method.

```python
from amltk.scheduling import Emitter, Event
emitter = Emitter("my-emitter")

class GPT:

    event: Event[str] = Event("my-event")

    def __init__(self) -> None:
        self.on_answer: Subscriber[str] = emitter.subscriber(self.event)

    def ask(self, question: str) -> None:
        emitter.emit(self.event, "hello world!")

gpt = GPT()

@gpt.on_answer
def print_answer(answer: str) -> None:
    print(answer)

gpt.ask("What is the conical way for an AI to greet someone?")
```

Typically these event based systems make little sense in a synchronous context, however
with the [`Scheduler`][amltk.scheduling.Scheduler] and [`Task`][amltk.scheduling.Task]
classes, they are used to enable a simple way to use multiprocessing and remote compute.
"""  # noqa: E501
from __future__ import annotations

import logging
import math
import time
from collections import Counter, defaultdict
from collections.abc import Callable, Iterable, Iterator, Mapping
from dataclasses import dataclass, field
from functools import partial
from typing import TYPE_CHECKING, Any, Generic, TypeVar, overload
from typing_extensions import ParamSpec, override
from uuid import uuid4

from amltk._functional import callstring, funcname
from amltk._richutil.renderers.function import Function

if TYPE_CHECKING:
    from rich.text import Text
    from rich.tree import Tree

T = TypeVar("T")
P = ParamSpec("P")
R = TypeVar("R")

logger = logging.getLogger(__name__)


@dataclass
class RegisteredTimeCallOrderStrategy:
    """A calling strategy that calls callbacks in the order they were registered."""

    @classmethod
    def execute(
        cls,
        events: Iterable[
            tuple[
                Iterable[Handler[P]],
                tuple[Any, ...] | None,
                dict[str, Any] | None,
            ]
        ],
    ) -> None:
        """Call all events in the scheduler."""
        all_handlers = []
        for handlers, args, kwargs in events:
            all_handlers += [
                (handler, args or (), kwargs or {}) for handler in handlers
            ]

        sorted_handlers = sorted(all_handlers, key=lambda item: item[0].registered_at)
        for handler, args, kwargs in sorted_handlers:
            handler(*args, **kwargs)


@dataclass(frozen=True)
class Event(Generic[P]):
    """An event that can be emitted.

    Attributes:
        name: The name of the event.
    """

    name: str

    @override
    def __eq__(self, other: Any) -> bool:
        """Check if two events are equal."""
        if isinstance(other, str):
            return self.name == other
        if isinstance(other, Event):
            return self.name == other.name

        return False

    def __rich__(self) -> Text:
        from rich.text import Text

        return Text("@", style="magenta bold").append(
            self.name,
            style="magenta italic",
        )


@dataclass
class Subscriber(Generic[P]):
    """An object that can be used to easily subscribe to a certain event.

    ```python
    from amltk.scheduling.events import Event, Subscriber

    test_event: Event[[int, str]] = Event("test")

    emitter = Emitter("hello world")
    subscribe = emitter.subscriber(test_event)

    # Use it as a decorator

    @subscribe
    def callback(a: int, b: str) -> None:
        print(f"Got {a} and {b}!")

    # ... or just pass a function

    subscribe(callback)

    # Will emit `test_event` with the arguments 1 and "hello"
    # and call the callback with those same arguments.
    emitter.emit(test_event, 1, "hello")
    ```

    Attributes:
        manager: The event manager to use.
        event: The event to subscribe to.
    """

    emitter: Emitter
    event: Event[P]
    when: Callable[[], bool] | None = None
    max_calls: int | None = None
    repeat: int = 1
    every: int = 1

    @property
    def event_counts(self) -> int:
        """The number of times this event has been emitted."""
        return self.emitter.event_counts[self.event]

    @overload
    def __call__(
        self,
        callback: None = None,
        *,
        when: Callable[[], bool] | None = ...,
        max_calls: int | None = ...,
        repeat: int = ...,
        every: int = ...,
    ) -> partial[Callable[P, Any]]:
        ...

    @overload
    def __call__(
        self,
        callback: Callable[P, Any],
        *,
        when: Callable[[], bool] | None = ...,
        max_calls: int | None = ...,
        repeat: int = ...,
        every: int = ...,
        hidden: bool = ...,
    ) -> Callable[P, Any]:
        ...

    def __call__(
        self,
        callback: Callable[P, Any] | None = None,
        *,
        when: Callable[[], bool] | None = None,
        max_calls: int | None = None,
        repeat: int = 1,
        every: int = 1,
        hidden: bool = False,
    ) -> Callable[P, Any] | partial[Callable[P, Any]]:
        """Subscribe to the event associated with this object.

        Args:
            callback: The callback to register.
            when: A predicate that must be satisfied for the callback to be called.
            every: The callback will be called every `every` times the event is emitted.
            repeat: The callback will be called `repeat` times successively.
            max_calls: The maximum number of times the callback can be called.
            hidden: Whether to hide the callback in visual output.
                This is mainly used to facilitate Plugins who
                act upon events but don't want to be seen, primarily
                as they are just book-keeping callbacks.

        Returns:
            The callback if it was provided, otherwise it acts
                as a decorator.
        """
        if callback is None:
            return partial(
                self.__call__,
                when=when,
                max_calls=max_calls,
                repeat=repeat,
                every=every,
            )  # type: ignore

        self.emitter.on(
            self.event,
            callback,
            when=when,
            max_calls=max_calls,
            repeat=repeat,
            every=every,
            hidden=hidden,
        )
        return callback

    def emit(self, *args: P.args, **kwargs: P.kwargs) -> None:
        """Emit this subscribers event."""
        self.emitter.emit(self.event, *args, **kwargs)


@dataclass
class Handler(Generic[P]):
    """A handler for an event.

    This is a simple class that holds a callback and any predicate
    that must be satisfied for it to be triggered.
    """

    callback: Callable[P, Any]
    when: Callable[[], bool] | None = None
    every: int = 1
    n_calls_to_handler: int = 0
    n_calls_to_callback: int = 0
    max_calls: int | None = None
    repeat: int = 1
    registered_at: int = field(default_factory=time.time_ns)
    hidden: bool = False

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> None:
        """Call the callback if the predicate is satisfied.

        If the predicate is not satisfied, then `None` is returned.
        """
        self.n_calls_to_handler += 1
        if self.every > 1 and self.n_calls_to_handler % self.every != 0:
            return

        if self.when is not None and not self.when():
            return

        max_calls = self.max_calls if self.max_calls is not None else math.inf
        for _ in range(self.repeat):
            if self.n_calls_to_callback >= max_calls:
                return

            logger.debug(f"Calling: {callstring(self.callback)}")
            self.callback(*args, **kwargs)
            self.n_calls_to_callback += 1

    def __rich__(self) -> Text:
        from rich.text import Text

        f_rich = Function(self.callback).__rich__()
        if self.n_calls_to_callback == 0:
            return f_rich

        return Text.assemble(
            f_rich,
            Text(" (", style="italic"),
            Text(f"{self.n_calls_to_callback}", style="yellow italic"),
            Text(")", style="italic"),
        )


class Emitter(Mapping[Event, list[Handler]]):
    """An event emitter.

    This class is used to emit events and register callbacks for those events.
    It also provides a convenience function
    [`subscriber()`][amltk.scheduling.events.Emitter.subscriber] such
    that objects using an `Emitter` can easily create access points for users
    to directly subscribe to their [`Events`][amltk.scheduling.events.Event].
    """

    name: str | None
    """The name of the emitter."""

    handlers: dict[Event, list[Handler]]
    """A mapping of events to their handlers."""

    event_counts: Counter[Event]
    """A count of all events emitted by this emitter."""

    def __init__(self, name: str | None = None) -> None:
        """Initialise the emitter.

        Args:
            name: The name of the emitter. If not provided, then a UUID
                will be used.
        """
        super().__init__()
        self.unique_ref = f"{name}-{uuid4()}"
        self.emitted_events: set[Event] = set()

        self.name = name
        self.handlers = defaultdict(list)
        self.event_counts = Counter()

    def emit(self, event: Event[P], *args: Any, **kwargs: Any) -> None:
        """Emit an event.

        This will call all the handlers for the event.

        Args:
            event: The event to emit.
                If passing a list, then the handlers for all events will be called,
                regardless of the order
            *args: The positional arguments to pass to the handlers.
            **kwargs: The keyword arguments to pass to the handlers.

        Returns:
            A list of the results from the handlers.
        """
        logger.debug(f"{self.name}: Emitting {event}")

        self.event_counts[event] += 1
        for handler in self.handlers[event]:
            handler(*args, **kwargs)

    def emit_many(
        self,
        events: dict[Event, tuple[tuple[Any, ...] | None, dict[str, Any] | None]],
    ) -> None:
        """Emit multiple events.

        This is useful for cases where you don't want to favour one callback
        over another, and so uses the time a callback was registered to call
        the callback instead.

        Args:
            events: A mapping of event keys to a tuple of positional
                arguments and keyword arguments to pass to the handlers.
        """
        for event in events:
            self.event_counts[event] += 1

        items = [
            (handlers, args, kwargs)
            for event, (args, kwargs) in events.items()
            if (handlers := self.get(event)) is not None
        ]

        header = f"{self.name}: Emitting many events"
        logger.debug(header)
        logger.debug(",".join(str(event) for event in events))
        RegisteredTimeCallOrderStrategy.execute(items)

    @override
    def __getitem__(self, event: Event) -> list[Handler]:
        return self.handlers[event]

    @override
    def __iter__(self) -> Iterator[Event]:
        return iter(self.handlers)

    @override
    def __len__(self) -> int:
        return len(self.handlers)

    @property
    def events(self) -> list[Event]:
        """Return a list of the events."""
        return list(self.handlers.keys())

    def subscriber(
        self,
        event: Event[P],
        *,
        when: Callable[[], bool] | None = None,
        every: int = 1,
        repeat: int = 1,
        max_calls: int | None = None,
    ) -> Subscriber[P]:
        """Create a subscriber for an event.

        Args:
            event: The event to register the callback for.
            when: A predicate that must be satisfied for the callback to be called.
            every: The callback will be called every `every` times the event is emitted.
            repeat: The callback will be called `repeat` times successively.
            max_calls: The maximum number of times the callback can be called.
        """
        if event not in self.handlers:
            self.handlers[event] = []

        return Subscriber(
            self,
            event,  # type: ignore
            when=when,
            every=every,
            repeat=repeat,
            max_calls=max_calls,
        )

    def on(
        self,
        event: Event[P],
        callback: Callable,
        *,
        when: Callable[[], bool] | None = None,
        every: int = 1,
        repeat: int = 1,
        max_calls: int | None = None,
        hidden: bool = False,
    ) -> None:
        """Register a callback for an event.

        Args:
            event: The event to register the callback for.
            callback: The callback to register.
            when: A predicate that must be satisfied for the callback to be called.
            every: The callback will be called every `every` times the event is emitted.
            repeat: The callback will be called `repeat` times successively.
            max_calls: The maximum number of times the callback can be called.
            hidden: Whether to hide the callback in visual output.
                This is mainly used to facilitate Plugins who
                act upon events but don't want to be seen, primarily
                as they are just book-keeping callbacks.
        """
        if repeat <= 0:
            raise ValueError(f"{repeat=} must be a positive integer.")

        if every <= 0:
            raise ValueError(f"{every=} must be a positive integer.")

        # Make sure it shows up in the event counts, setting it to 0 if it
        # doesn't exist
        self.event_counts.setdefault(event, 0)

        # This hackery is just to get down to a flat list of events that need
        # to be set up
        self.handlers[event].append(
            Handler(
                callback,
                when=when,
                every=every,
                repeat=repeat,
                max_calls=max_calls,
                hidden=hidden,
            ),
        )

        _name = funcname(callback)
        msg = f"{self.name}: Registered callback '{_name}' for event {event}"
        if every > 1:
            msg += f" every {every} times"
        if when:
            msg += f" with predicate ({funcname(when)})"
        if repeat > 1:
            msg += f" called {repeat} times successively"
        if hidden:
            msg += " (hidden from visual output)"
        logger.debug(msg)

    def add_event(self, *event: Event) -> None:
        """Add an event to the event manager so that it shows up in visuals.

        Args:
            event: The event to add.
        """
        for e in event:
            if e not in self.handlers:
                self.handlers[e] = []

    def __rich__(self) -> Tree:
        from rich.tree import Tree

        tree = Tree(self.name or "", hide_root=self.name is None)

        # This just groups events with callbacks together
        handler_items = sorted(
            self.handlers.items(),
            key=lambda item: not any(item[1])
            or not all(handler.hidden for handler in item[1]),
        )

        for event, _handlers in handler_items:
            event_text = event.__rich__()
            if (count := self.event_counts[event]) >= 1:
                event_text.append(f" {count}", style="yellow italic")

            event_tree = tree.add(event_text)
            for handler in _handlers:
                if not handler.hidden:
                    event_tree.add(handler)

        return tree



"""A [`QueueMonitor`][amltk.scheduling.queue_monitor.QueueMonitor] is a
monitor for the scheduler queue.

This module contains a monitor for the scheduler queue. The monitor tracks the
queue state at every event emitted by the scheduler. The data can be converted
to a pandas DataFrame or plotted as a stacked barchart.

!!! note "Monitoring Frequency"

    To prevent repeated polling, we sample the scheduler queue at every scheduler event.
    This is because the queue is only modified upon one of these events. This means we
    don't need to poll the queue at a fixed interval. However, if you need more fine
    grained updates, you can add extra events/timings at which the monitor should
    [`update()`][amltk.scheduling.queue_monitor.QueueMonitor.update].

!!! warning "Performance impact"

    If your tasks and callbacks are very fast (~sub 10ms), then the monitor has a
    non-nelgible impact however for most use cases, this should not be a problem.
    As anything, you should profile how much work the scheduler can get done,
    with and without the monitor, to see if it is a problem for your use case.

In the below example, we have a very fast running function that runs on repeat,
sometimes too fast for the scheduler to keep up, letting some futures buildup needing
to be processed.

```python exec="true" source="material-block" result="python" session="queue-monitor"
import time
import matplotlib.pyplot as plt
from amltk.scheduling import Scheduler
from amltk.scheduling.queue_monitor import QueueMonitor

def fast_function(x: int) -> int:
    return x + 1
from amltk._doc import make_picklable; make_picklable(fast_function)  # markdown-exec: hide

N_WORKERS = 2
scheduler = Scheduler.with_processes(N_WORKERS)
monitor = QueueMonitor(scheduler)
task = scheduler.task(fast_function)

@scheduler.on_start(repeat=N_WORKERS)
def start():
    task.submit(1)

@task.on_result
def result(_, x: int):
    if scheduler.running():
        task.submit(x)

scheduler.run(timeout=1)
df = monitor.df()
print(df)
```

We can also [`plot()`][amltk.scheduling.queue_monitor.QueueMonitor.plot] the data as a
stacked barchart with a set interval.

```python exec="true" source="material-block" html="true" session="queue-monitor"
fig, ax = plt.subplots()
monitor.plot(interval=(50, "ms"))
from io import StringIO; fig.tight_layout(); buffer = StringIO(); plt.savefig(buffer, format="svg"); print(buffer.getvalue())  # markdown-exec: hide
```

"""  # noqa: E501
from __future__ import annotations

import time
from collections import Counter
from typing import TYPE_CHECKING, Any, NamedTuple

import matplotlib.pyplot as plt
import pandas as pd
from matplotlib.ticker import MaxNLocator

if TYPE_CHECKING:
    from pandas.core.tools.datetimes import UnitChoices

    from amltk.scheduling import Scheduler


class QueueMonitorRecord(NamedTuple):
    """A record of the queue state at a given time."""

    time: int  # recorded in time_nanoseconds
    queue_size: int
    queued: int
    finished: int
    cancelled: int


class QueueMonitor:
    """A monitor for the scheduler queue."""

    def __init__(self, scheduler: Scheduler) -> None:
        """Initializes the monitor."""
        super().__init__()
        self.scheduler = scheduler
        self.data: list[QueueMonitorRecord] = []

        scheduler.on_start(self.update)
        scheduler.on_finishing(self.update)
        scheduler.on_finished(self.update)
        scheduler.on_future_submitted(self.update)
        scheduler.on_future_cancelled(self.update)
        scheduler.on_future_done(self.update)

    def df(
        self,
        *,
        n_workers: int | None = None,
    ) -> pd.DataFrame:
        """Converts the data to a pandas DataFrame.

        Args:
            n_workers: The number of workers that were in use. This helps idenify how
                many workers were idle at a given time. If None, the maximum length of
                the queue at any recorded time is used.
        """
        _df = pd.DataFrame(self.data, columns=list(QueueMonitorRecord._fields)).astype(
            {
                # Windows might have a weird default here but it should be 64 at least
                "time": "int64",
                "queue_size": int,
                "queued": int,
                "finished": int,
                "cancelled": int,
            },
        )
        if n_workers is None:
            n_workers = int(_df["queue_size"].max())
        _df["idle"] = n_workers - _df["queue_size"]
        _df["time"] = pd.to_datetime(_df["time"], unit="ns", origin="unix")
        return _df.set_index("time")

    def plot(
        self,
        *,
        ax: plt.Axes | None = None,
        interval: tuple[int, UnitChoices] = (1, "s"),
        n_workers: int | None = None,
        **kwargs: Any,
    ) -> plt.Axes:
        """Plots the data as a stacked barchart.

        Args:
            ax: The axes to plot on. If None, a new figure is created.
            interval: The interval to use for the x-axis. The first value is the
                interval and the second value is the unit. Must be a valid pandas
                timedelta unit. See [to_timedelta()][pandas.to_timedelta] for more
                information.
            n_workers: The number of workers that were in use. This helps idenify how
                many workers were idle at a given time. If None, the maximum length of
                the queue at any recorded time is used.
            **kwargs: Additional keyword arguments to pass to the pandas plot function.

        Returns:
            The axes.
        """
        if ax is None:
            _, _ax = plt.subplots(1, 1)
        else:
            _ax = ax

        _df = self.df(n_workers=n_workers)
        _df = _df.resample(f"{interval[0]}{interval[1]}").mean()
        _df.index = _df.index - _df.index[0]
        _reversed_df = _df[::-1]

        _reversed_df.plot.barh(
            stacked=True,
            y=["finished", "queued", "cancelled", "idle"],
            ax=_ax,
            width=1,
            edgecolor="k",
            **kwargs,
        )

        _ax.set_ylabel("Time")
        _ax.yaxis.set_major_locator(MaxNLocator(nbins="auto"))

        _ax.set_xlabel("Count")
        _ax.xaxis.set_major_locator(MaxNLocator(integer=True))

        return _ax

    def update(self, *_: Any) -> None:
        """Updates the data when the scheduler has an event."""
        queue = self.scheduler.queue
        # OPTIM: Not sure if this is fastenough
        counter = Counter([f._state for f in queue])
        self.data.append(
            QueueMonitorRecord(
                time.time_ns(),
                len(queue),
                counter["PENDING"],
                counter["FINISHED"],
                counter["CANCELLED"],
            ),
        )



from amltk.scheduling.plugins.comm import Comm
from amltk.scheduling.plugins.limiter import Limiter
from amltk.scheduling.plugins.plugin import Plugin
from amltk.scheduling.plugins.warning_filter import WarningFilter

__all__ = [
    "Limiter",
    "WarningFilter",
    "Plugin",
    "Comm",
]



"""Loaders for [`PathBucket`][amltk.store.paths.path_bucket.PathBucket]s.

The [`Loader`][amltk.store.paths.path_loaders.PathLoader]s in this module are used to
load and save objects identified by a unique [`Path`][pathlib.Path].
For saving objects, these loaders rely on checking the type of the
object for [`can_save`][amltk.store.paths.path_loaders.PathLoader.can_save] and
[`save`][amltk.store.paths.path_loaders.PathLoader.save] methods. For loading objects,
these loaders rely on checking the file extension of the path for
[`can_load`][amltk.store.paths.path_loaders.PathLoader.can_load] and
[`load`][amltk.store.paths.path_loaders.PathLoader.load] methods.
"""
from __future__ import annotations

import json
import logging
import pickle
from abc import abstractmethod
from pathlib import Path
from typing import TYPE_CHECKING, Any, ClassVar, TypeVar
from typing_extensions import override

import numpy as np
import pandas as pd

from amltk.store.loader import Loader

if TYPE_CHECKING:
    from types import ModuleType

# NOTE: Since we don't want to depend on yaml, we do a dynamic
# import on it.
yaml: ModuleType | None
try:
    import yaml
except ImportError:
    yaml = None

T = TypeVar("T")

logger = logging.getLogger(__name__)


class PathLoader(Loader[Path, T]):
    """A [`Loader`][amltk.store.loader.Loader] for loading and saving
    objects indentified by a [`Path`][pathlib.Path].
    """

    name: ClassVar[str]
    """The name of the loader."""

    @override
    @classmethod
    @abstractmethod
    def can_load(cls, key: Path, /, *, check: type[T] | None = None) -> bool:
        """Return True if this loader supports the resource at key.

        This is used to determine which loader to use when loading a
        resource from a key.

        Args:
            key: The key used to identify the resource
            check: If the loader can support loading a specific type
                of object.
        """
        ...

    @override
    @classmethod
    @abstractmethod
    def can_save(cls, obj: Any, key: Path, /) -> bool:
        """Return True if this loader can save this object.

        This is used to determine which loader to use when loading a
        resource from a key.

        Args:
            obj: The object to save.
            key: The key used to identify the resource
        """
        ...

    @override
    @classmethod
    @abstractmethod
    def save(cls, obj: Any, key: Path, /) -> None:
        """Save an object to under the given key.

        Args:
            obj: The object to save.
            key: The key to save the object under.
        """
        ...

    @override
    @classmethod
    @abstractmethod
    def load(cls, key: Path, /) -> T:
        """Load an object from the given key.

        Args:
            key: The key to load the object from.

        Returns:
            The loaded object.
        """
        ...


class NPYLoader(PathLoader[np.ndarray]):
    """A [`Loader`][amltk.store.paths.path_loaders.PathLoader] for loading and
    saving [`np.ndarray`][numpy.ndarray]s.

    This loader supports the following file extensions:

    * `#!python ".npy"`

    This loader supports the following types:

    * [`np.ndarray`][numpy.ndarray]
    """

    name: ClassVar = "np"
    """::: amltk.store.paths.path_loaders.PathLoader.name"""

    @override
    @classmethod
    def can_save(cls, obj: Any, key: Path, /) -> bool:
        """::: amltk.store.paths.path_loaders.PathLoader.can_save"""  # noqa: D415
        return isinstance(obj, np.ndarray) and key.suffix in {".npy"}

    @override
    @classmethod
    def can_load(cls, key: Path, /, *, check: type | None = None) -> bool:
        """::: amltk.store.paths.path_loaders.PathLoader.can_load"""  # noqa: D415
        return key.suffix in {".npy"} and check in (np.ndarray, None)

    @override
    @classmethod
    def load(cls, key: Path, /) -> np.ndarray:
        """::: amltk.store.paths.path_loaders.PathLoader.load"""  # noqa: D415
        logger.debug(f"Loading {key=}")
        item = np.load(key, allow_pickle=False)
        if not isinstance(item, np.ndarray):
            msg = f"Expected `np.ndarray` from {key=} but got `{type(item).__name__}`."
            raise TypeError(msg)

        return item

    @override
    @classmethod
    def save(cls, obj: np.ndarray, key: Path, /) -> None:
        """::: amltk.store.paths.path_loaders.PathLoader.save"""  # noqa: D415
        logger.debug(f"Saving {key=}")
        np.save(key, obj, allow_pickle=False)


class PDLoader(PathLoader[pd.DataFrame | pd.Series]):
    """A [`Loader`][amltk.store.paths.path_loaders.PathLoader] for loading and
    saving [`pd.DataFrame`][pandas.DataFrame]s.

    This loader supports the following file extensions:

    * `#!python ".csv"`
    * `#!python ".parquet"`
    * `#!python ".pdpickle"`

    This loader supports the following types:

    * [`pd.DataFrame`][pandas.DataFrame]
    * [`pd.Series`][pandas.Series] - Only to `#!python ".pdpickle"` files

    ???+ note "Multiindex support"

        There is currently no multi-index support as we explicitly
        use `index_col=0` when loading a `".csv"` file. This is
        because we assume that the first column is the index to
        prevent Unamed columns from being created.

    ???+ note "Series support"

        There is currently limited support for pandas series as once written
        to csv/parquet, they are converted to a dataframe with a single column.
        See [this issue](https://github.com/automl/amltk/issues/4)

        Please consider using `".pdpickle"` instead.
    """

    name: ClassVar = "pd"
    """::: amltk.store.paths.path_loaders.PathLoader.name"""

    @override
    @classmethod
    def can_load(cls, key: Path, /, *, check: type | None = None) -> bool:
        """::: amltk.store.paths.path_loaders.PathLoader.can_load"""  # noqa: D415
        if key.suffix in (".pdpickle", None):
            return check in (pd.Series, pd.DataFrame, None)

        if key.suffix in (".csv", ".parquet"):
            return check in (pd.DataFrame, None)

        return False

    @override
    @classmethod
    def can_save(cls, obj: Any, key: Path, /) -> bool:
        """::: amltk.store.paths.path_loaders.PathLoader.can_save"""  # noqa: D415
        if key.suffix == ".pdpickle":
            return isinstance(obj, pd.Series | pd.DataFrame)

        if key.suffix == ".parquet":
            return isinstance(obj, pd.DataFrame)

        if key.suffix == ".csv":
            # TODO: https://github.com/automl/amltk/issues/4
            return isinstance(obj, pd.DataFrame) and obj.index.nlevels == 1

        return False

    @override
    @classmethod
    def load(cls, key: Path, /) -> pd.DataFrame | pd.Series:
        """::: amltk.store.paths.path_loaders.PathLoader.load"""  # noqa: D415
        logger.debug(f"Loading {key=}")
        if key.suffix == ".csv":
            return pd.read_csv(key, index_col=0)

        if key.suffix == ".parquet":
            return pd.read_parquet(key)

        if key.suffix == ".pdpickle":
            obj = pd.read_pickle(key)  # noqa: S301
            if not isinstance(obj, pd.Series | pd.DataFrame):
                msg = (
                    f"Expected `pd.Series | pd.DataFrame` from {key=}"
                    f" but got `{type(obj).__name__}`."
                )
                raise TypeError(msg)

            return obj

        raise ValueError(f"Unknown file extension {key.suffix}")

    @override
    @classmethod
    def save(cls, obj: pd.Series | pd.DataFrame, key: Path, /) -> None:
        """::: amltk.store.paths.path_loaders.PathLoader.save"""  # noqa: D415
        # Most pandas methods only seem to support dataframes
        logger.debug(f"Saving {key=}")

        if key.suffix == ".pdpickle":
            obj.to_pickle(key)
            return

        if key.suffix == ".csv":
            if obj.index.name is None and obj.index.nlevels == 1:
                obj.index.name = "index"

            obj.to_csv(key, index=True)
            return

        if key.suffix == ".parquet":
            obj.to_parquet(key)
            return

        raise ValueError(f"Unknown extension {key.suffix=}")


class JSONLoader(PathLoader[dict | list]):
    """A [`Loader`][amltk.store.paths.path_loaders.PathLoader] for loading and
    saving [`dict`][dict]s and [`list`][list]s to JSON.

    This loader supports the following file extensions:

    * `#!python ".json"`

    This loader supports the following types:

    * [`dict`][dict]
    * [`list`][list]
    """

    name: ClassVar = "json"
    """::: amltk.store.paths.path_loaders.PathLoader.name"""

    @override
    @classmethod
    def can_load(cls, key: Path, /, *, check: type | None = None) -> bool:
        """::: amltk.store.paths.path_loaders.PathLoader.can_load"""  # noqa: D415
        return key.suffix == ".json" and check in (dict, list, None)

    @override
    @classmethod
    def can_save(cls, obj: Any, key: Path, /) -> bool:
        """::: amltk.store.paths.path_loaders.PathLoader.can_save"""  # noqa: D415
        return isinstance(obj, dict | list) and key.suffix == ".json"

    @override
    @classmethod
    def load(cls, key: Path, /) -> dict | list:
        """::: amltk.store.paths.path_loaders.PathLoader.load"""  # noqa: D415
        logger.debug(f"Loading {key=}")
        with key.open("r") as f:
            item = json.load(f)

        if not isinstance(item, dict | list):
            msg = f"Expected `dict | list` from {key=} but got `{type(item).__name__}`"
            raise TypeError(msg)

        return item

    @override
    @classmethod
    def save(cls, obj: dict | list, key: Path, /) -> None:
        """::: amltk.store.paths.path_loaders.PathLoader.save"""  # noqa: D415
        logger.debug(f"Saving {key=}")
        with key.open("w") as f:
            json.dump(obj, f)


class YAMLLoader(PathLoader[dict | list]):
    """A [`Loader`][amltk.store.paths.path_loaders.PathLoader] for loading and
    saving [`dict`][dict]s and [`list`][list]s to YAML.

    This loader supports the following file extensions:

    * `#!python ".yaml"`
    * `#!python ".yml"`

    This loader supports the following types:

    * [`dict`][dict]
    * [`list`][list]
    """

    name: ClassVar = "yaml"
    """::: amltk.store.paths.path_loaders.PathLoader.name"""

    @override
    @classmethod
    def can_load(cls, key: Path, /, *, check: type | None = None) -> bool:
        """::: amltk.store.paths.path_loaders.PathLoader.can_load"""  # noqa: D415
        return key.suffix in (".yaml", ".yml") and check in (dict, list, None)

    @override
    @classmethod
    def can_save(cls, obj: Any, key: Path, /) -> bool:
        """::: amltk.store.paths.path_loaders.PathLoader.can_save"""  # noqa: D415
        return isinstance(obj, dict | list) and key.suffix in (".yaml", ".yml")

    @override
    @classmethod
    def load(cls, key: Path, /) -> dict | list:
        """::: amltk.store.paths.path_loaders.PathLoader.load"""  # noqa: D415
        logger.debug(f"Loading {key=}")
        if yaml is None:
            raise ModuleNotFoundError("PyYAML is not installed")

        with key.open("r") as f:
            item = yaml.safe_load(f)

        if not isinstance(item, dict | list):
            msg = f"Expected `dict | list` from {key=} but got `{type(item).__name__}`"
            raise TypeError(msg)

        return item

    @override
    @classmethod
    def save(cls, obj: dict | list, key: Path, /) -> None:
        """::: amltk.store.paths.path_loaders.PathLoader.save"""  # noqa: D415
        logger.debug(f"Saving {key=}")
        if yaml is None:
            raise ModuleNotFoundError("PyYAML is not installed")

        with key.open("w") as f:
            yaml.dump(obj, f)


class PickleLoader(PathLoader[Any]):
    """A [`Loader`][amltk.store.paths.path_loaders.PathLoader] for loading and
    saving any object to a pickle file.

    This loader supports the following file extensions:

    * `#!python ".pkl"`

    This loader supports the following types:

    * object

    ???+ note "Picklability"
        This loader uses Python's built-in [`pickle`][pickle] module
        to save and load objects. This means that the object must be
        [picklable][pickle] in order to be saved and loaded. If the
        object is not picklable, then an error will be raised when
        attempting to save or load the object.
    """

    name: ClassVar = "pickle"
    """::: amltk.store.paths.path_loaders.PathLoader.name"""

    @override
    @classmethod
    def can_load(
        cls,
        key: Path,
        /,
        *,
        check: type | None = None,
    ) -> bool:
        """::: amltk.store.paths.path_loaders.PathLoader.can_load"""  # noqa: D415
        return key.suffix in (".pkl", ".pickle")

    @override
    @classmethod
    def can_save(cls, obj: Any, key: Path, /) -> bool:
        """::: amltk.store.paths.path_loaders.PathLoader.can_save"""  # noqa: D415
        return True  # Anything can be attempted to be pickled

    @override
    @classmethod
    def load(cls, key: Path, /) -> Any:
        """::: amltk.store.paths.path_loaders.PathLoader.load"""  # noqa: D415
        logger.debug(f"Loading {key=}")
        with key.open("rb") as f:
            return pickle.load(f)  # noqa: S301

    @override
    @classmethod
    def save(cls, obj: Any, key: Path, /) -> None:
        """::: amltk.store.paths.path_loaders.PathLoader.save"""  # noqa: D415
        logger.debug(f"Saving {key=}")
        with key.open("wb") as f:
            pickle.dump(obj, f)


class TxtLoader(PathLoader[str]):
    """A [`Loader`][amltk.store.paths.path_loaders.PathLoader] for loading and
    saving [`str`][str]s to text files.

    This loader supports the following file extensions:

    * `#!python ".text"`
    * `#!python ".txt"`

    This loader supports the following types:

    * [`str`][str]
    """

    name: ClassVar = "text"
    """::: amltk.store.paths.path_loaders.PathLoader.name"""

    @override
    @classmethod
    def can_load(cls, key: Path, /, *, check: type | None = None) -> bool:
        """::: amltk.store.paths.path_loaders.PathLoader.can_load"""  # noqa: D415
        return key.suffix in (".text", ".txt") and check in (str, None)

    @override
    @classmethod
    def can_save(cls, obj: Any, key: Path, /) -> bool:
        """::: amltk.store.paths.path_loaders.PathLoader.can_save"""  # noqa: D415
        return isinstance(obj, str) and key.suffix in (".text", ".txt")

    @override
    @classmethod
    def load(cls, key: Path, /) -> str:
        """::: amltk.store.paths.path_loaders.PathLoader.load"""  # noqa: D415
        logger.debug(f"Loading {key=}")
        with key.open("r") as f:
            return f.read()

    @override
    @classmethod
    def save(cls, obj: str, key: Path, /) -> None:
        """::: amltk.store.paths.path_loaders.PathLoader.save"""  # noqa: D415
        logger.debug(f"Saving {key=}")
        with key.open("w") as f:
            f.write(obj)


class ByteLoader(PathLoader[bytes]):
    """A [`Loader`][amltk.store.paths.path_loaders.PathLoader] for loading and
    saving [`bytes`][bytes] to binary files.

    This loader supports the following file extensions:

    * `#!python ".bin"`
    * `#!python ".bytes"`

    This loader supports the following types:

    * [`bytes`][bytes]
    """

    name: ClassVar = "bytes"

    @override
    @classmethod
    def can_load(cls, key: Path, /, *, check: type | None = None) -> bool:
        """::: amltk.store.paths.path_loaders.PathLoader.can_load"""  # noqa: D415
        return key.suffix in (".bin", ".bytes") and check in (bytes, None)

    @override
    @classmethod
    def can_save(cls, obj: Any, key: Path, /) -> bool:
        """::: amltk.store.paths.path_loaders.PathLoader.can_save"""  # noqa: D415
        return isinstance(obj, dict | list) and key.suffix in (".bin", ".bytes")

    @override
    @classmethod
    def load(cls, key: Path, /) -> bytes:
        """::: amltk.store.paths.path_loaders.PathLoader.load"""  # noqa: D415
        logger.debug(f"Loading {key=}")
        with key.open("rb") as f:
            return f.read()

    @override
    @classmethod
    def save(cls, obj: bytes, key: Path, /) -> None:
        """::: amltk.store.paths.path_loaders.PathLoader.save"""  # noqa: D415
        logger.debug(f"Saving {key=}")
        with key.open("wb") as f:
            f.write(obj)



from amltk.store.bucket import Bucket
from amltk.store.drop import Drop
from amltk.store.loader import Loader
from amltk.store.paths.path_bucket import PathBucket
from amltk.store.paths.path_loaders import (
    ByteLoader,
    JSONLoader,
    NPYLoader,
    PathLoader,
    PDLoader,
    PickleLoader,
    TxtLoader,
    YAMLLoader,
)
from amltk.store.stored_value import StoredValue

__all__ = [
    "Bucket",
    "Drop",
    "Loader",
    "PathBucket",
    "PDLoader",
    "NPYLoader",
    "JSONLoader",
    "YAMLLoader",
    "PickleLoader",
    "TxtLoader",
    "ByteLoader",
    "PathLoader",
    "StoredValue",
]



"""Emissions Tracker Plugin Module.

This module defines a plugin for tracking carbon emissions using the codecarbon library.

For usage examples, refer to the docstring of the EmissionsTrackerPlugin class.
"""
from __future__ import annotations

from collections.abc import Callable
from typing import TYPE_CHECKING, Any, ClassVar, Generic, TypeVar
from typing_extensions import ParamSpec, Self

from codecarbon import EmissionsTracker

from amltk.scheduling.plugins.plugin import Plugin

if TYPE_CHECKING:
    from rich.panel import Panel

    from amltk.scheduling.task import Task

P = ParamSpec("P")
R = TypeVar("R")


class _EmissionsTrackerWrapper(Generic[P, R]):
    """A wrapper around codecarbon package to measure emissions."""

    def __init__(
        self,
        fn: Callable[P, R],
        *codecarbon_args: Any,
        **codecarbon_kwargs: Any,
    ):
        """Initialize the wrapper.

        Args:
            fn: The function to wrap.
            *codecarbon_args: Arguments to pass to codecarbon EmissionsTracker.
            **codecarbon_kwargs: Keyword args to pass to codecarbon EmissionsTracker.
        """
        super().__init__()
        self.fn = fn
        self.codecarbon_args = codecarbon_args
        self.codecarbon_kwargs = codecarbon_kwargs

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> R:
        with EmissionsTracker(*self.codecarbon_args, **self.codecarbon_kwargs):
            return self.fn(*args, **kwargs)


class EmissionsTrackerPlugin(Plugin):
    """A plugin that tracks carbon emissions using codecarbon library.

    Usage Example:

    ```python
    from concurrent.futures import ThreadPoolExecutor
    from amltk.scheduling import Scheduler
    from amltk.scheduling.plugins.emissions_tracker_plugin import EmissionsTrackerPlugin

    def some_function(x: int) -> int:
        return x * 2

    executor = ThreadPoolExecutor(max_workers=1)

    # Create a Scheduler instance with the executor
    scheduler = Scheduler(executor=executor)

    # Create a task with the emissions tracker plugin
    task = scheduler.task(some_function, plugins=[
        # Pass any codecarbon parameters as args here
        EmissionsTrackerPlugin(log_level="info", save_to_file=False)
    ])

    @scheduler.on_start
    def on_start():
        task.submit(5)  # Submit any args here

    @task.on_submitted
    def on_submitted(future, *args, **kwargs):
        print(f"Task was submitted", future, args, kwargs)

    @task.on_done
    def on_done(future):
        # Result is the return value of the function
        print("Task done: ", future.result())

    scheduler.run()
    ```
    """

    name: ClassVar = "emissions-tracker"
    """The name of the plugin."""

    def __init__(self, *args: Any, **kwargs: Any):
        """Initialize the EmissionsTrackerPlugin.

        Args:
            *args: Additional arguments to pass to codecarbon library.
            **kwargs: Additional keyword arguments to pass to codecarbon library.

        You can pass any codecarbon parameters as args to EmissionsTrackerPlugin.
        Please refer to the official codecarbon documentation for more details:
        https://mlco2.github.io/codecarbon/parameters.html
        """
        super().__init__()
        self.task: Task | None = None
        self.codecarbon_args = args
        self.codecarbon_kwargs = kwargs

    def attach_task(self, task: Task) -> None:
        """Attach the plugin to a task."""
        self.task = task

    def pre_submit(
        self,
        fn: Callable[P, R],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> tuple[Callable[P, R], tuple, dict]:
        """Pre-submit hook."""
        wrapped_f = _EmissionsTrackerWrapper(
            fn,
            self.task,
            *self.codecarbon_args,
            **self.codecarbon_kwargs,
        )
        return wrapped_f, args, kwargs

    def copy(self) -> Self:
        """Return a copy of the plugin."""
        return self.__class__(*self.codecarbon_args, **self.codecarbon_kwargs)

    def __rich__(self) -> Panel:
        """Return a rich panel."""
        from rich.panel import Panel

        return Panel(
            f"codecarbon_args: {self.codecarbon_args} "
            f"codecarbon_kwargs: {self.codecarbon_kwargs}",
            title=f"Plugin {self.name}",
        )



"""A module to hold some async specific functionality."""
from __future__ import annotations

import asyncio
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any
from typing_extensions import override

if TYPE_CHECKING:
    from multiprocessing.connection import Connection


@dataclass
class AsyncConnection:
    """A wrapper around a multiprocessing connection to make it async."""

    connection: Connection
    """The underlying connection"""

    async def recv(self) -> Any:
        """Receive a message from the connection.

        Returns:
            The received message.
        """
        is_readable = asyncio.Event()
        loop = asyncio.get_running_loop()
        loop.add_reader(self.connection.fileno(), is_readable.set)

        if not self.connection.poll():
            await is_readable.wait()

        result = self.connection.recv()
        is_readable.clear()
        loop.remove_reader(self.connection.fileno())
        return result

    async def send(self, obj: Any) -> None:
        """Send a message to the connection.

        Args:
            obj: The object to send.
        """
        is_writable = asyncio.Event()
        loop = asyncio.get_running_loop()
        loop.add_writer(self.connection.fileno(), is_writable.set)

        await is_writable.wait()

        self.connection.send(obj)
        is_writable.clear()
        loop.remove_writer(self.connection.fileno())


class ContextEvent(asyncio.Event):
    """An event with added context to why it was triggered.

    Attributes:
        msg: The message that was set.
        exception: The exception that was set.
    """

    def __init__(self, **kwargs: Any) -> None:
        """See [asyncio.Event][] for more information."""
        super().__init__(**kwargs)
        self.msg: str | None = None
        self.exception: BaseException | None = None

    @override
    def set(
        self,
        msg: str | None = None,
        exception: BaseException | None = None,
    ) -> None:
        """Set the event and set the context.

        Args:
            msg: The message to set.
            exception: The exception to set.
        """
        self.msg = msg
        self.exception = exception
        super().set()

    @override
    def clear(self) -> None:
        """Clear the event and clear the context."""
        self.msg = None
        self.exception = None
        super().clear()

    @property
    def context(self) -> tuple[str | None, BaseException | None]:
        """Get the context information.

        Returns:
            A tuple of the message and exception.
        """
        return self.msg, self.exception

    def __enter__(self) -> ContextEvent:
        """Enter the context manager."""
        self.set()
        return self

    def __exit__(self, *args: Any) -> None:
        """Exit the context manager."""
        self.clear()



"""Whether for debugging, building an AutoML system or for optimization
purposes, we provide a powerful [`Profiler`][amltk.profiling.Profiler],
which can generate a [`Profile`][amltk.profiling.Profile] of different sections
of code. This is particularly useful with [`Trial`][amltk.optimization.Trial]s,
so much so that we attach one to every `Trial` made as
[`trial.profiler`][amltk.optimization.Trial.profiler].

When done profiling, you can export all generated profiles as a dataframe using
[`profiler.df()`][amltk.profiling.Profiler.df].

```python exec="true" result="python" source="material-block"
from amltk.profiling import Profiler
import numpy as np

profiler = Profiler()

with profiler("loading-data"):
    X = np.random.rand(1000, 1000)

with profiler("training-model"):
    model = np.linalg.inv(X)

with profiler("predicting"):
    y = model @ X

print(profiler.df())
```

You'll find these profiles as keys in the [`Profiler`][amltk.profiling.Profiler],
e.g. `#! python profiler["loading-data"]`.

This will measure both the time it took within the block but also
the memory consumed before and after the block finishes, allowing
you to get an estimate of the memory consumed.


??? tip "Memory, vms vs rms"

    While not entirely accurate, this should be enough for info
    for most use cases.

    Given the main process uses 2GB of memory and the process
    then spawns a new process in which you are profiling, as you
    might do from a [`Task`][amltk.scheduling.Task]. In this new
    process you use another 2GB on top of that, then:

    * The virtual memory size (**vms**) will show 4GB as the
    new process will share the 2GB with the main process and
    have it's own 2GB.

    * The resident set size (**rss**) will show 2GB as the
    new process will only have 2GB of it's own memory.


If you need to profile some iterator, like a for loop, you can use
[`Profiler.each()`][amltk.profiling.Profiler.each] which will measure
the entire loop but also each individual iteration. This can be useful
for iterating batches of a deep-learning model, splits of a cross-validator
or really any loop with work you want to profile.

```python exec="true" result="python" source="material-block"
from amltk.profiling import Profiler
import numpy as np

profiler = Profiler()

for i in profiler.each(range(3), name="for-loop"):
    X = np.random.rand(1000, 1000)

print(profiler.df())
```

Lastly, to disable profiling without editing much code,
you can always use [`Profiler.disable()`][amltk.profiling.Profiler.disable]
and [`Profiler.enable()`][amltk.profiling.Profiler.enable] to toggle
profiling on and off.
"""

from __future__ import annotations

from collections import deque
from collections.abc import Callable, Iterable, Iterator, Mapping
from contextlib import contextmanager
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Literal, TypeVar
from typing_extensions import override

import pandas as pd

from amltk._functional import mapping_select
from amltk.profiling.memory import Memory
from amltk.profiling.timing import Timer

if TYPE_CHECKING:
    from rich.console import RenderableType

T = TypeVar("T")


@dataclass
class Profile:
    """A profiler for measuring statistics between two events."""

    timer: Timer
    memory: Memory

    @classmethod
    def from_dict(cls, d: Mapping[str, Any]) -> Profile.Interval:
        """Create a profile interval from a dictionary.

        Args:
            d: The dictionary to create from.

        Returns:
            The profile interval.
        """
        return Profile.Interval(
            memory=Memory.from_dict(mapping_select(d, "memory:")),
            time=Timer.from_dict(mapping_select(d, "time:")),
        )

    @classmethod
    @contextmanager
    def measure(
        cls,
        *,
        memory_unit: Memory.Unit | Literal["B", "KB", "MB", "GB"] = "B",
        time_kind: Timer.Kind | Literal["wall", "cpu", "process"] = "wall",
    ) -> Iterator[Profile.Interval]:
        """Profile a block of code.

        !!! note

            * See [`Memory`][amltk.profiling.Memory] for more information on memory.
            * See [`Timer`][amltk.profiling.Timer] for more information on timing.

        Args:
            memory_unit: The unit of memory to use.
            time_kind: The type of timer to use.

        Yields:
            The Profiler Interval. Memory and Timings will not be valid until
            the context manager is exited.
        """
        with Memory.measure(unit=memory_unit) as memory, Timer.time(
            kind=time_kind,
        ) as timer:
            yield Profile.Interval(memory=memory, time=timer)

    @classmethod
    def start(
        cls,
        memory_unit: Memory.Unit | Literal["B", "KB", "MB", "GB"] = "B",
        time_kind: Timer.Kind | Literal["wall", "cpu", "process"] = "wall",
    ) -> Profile:
        """Start a memory tracker.

        !!! note

            * See [`Memory`][amltk.profiling.Memory] for more information on memory.
            * See [`Timer`][amltk.profiling.Timer] for more information on timing.

        Args:
            memory_unit: The unit of memory to use.
            time_kind: The type of timer to use.

        Returns:
            The Memory tracker.
        """
        return Profile(
            timer=Timer.start(kind=time_kind),
            memory=Memory.start(unit=memory_unit),
        )

    def stop(self) -> Profile.Interval:
        """Stop the memory tracker.

        Returns:
            The memory interval.
        """
        return Profile.Interval(
            memory=self.memory.stop(),
            time=self.timer.stop(),
        )

    @classmethod
    def na(cls) -> Profile.Interval:
        """Create a profile interval that represents NA."""
        return Profile.Interval(memory=Memory.na(), time=Timer.na())

    @dataclass
    class Interval:
        """A class for representing a profiled interval."""

        memory: Memory.Interval
        time: Timer.Interval

        def to_dict(self, *, prefix: str = "") -> dict[str, Any]:
            """Convert the profile interval to a dictionary."""
            _prefix = "" if prefix == "" else f"{prefix}:"
            return {
                **self.memory.to_dict(prefix=f"{_prefix}memory:"),
                **self.time.to_dict(prefix=f"{_prefix}time:"),
            }


@dataclass
class Profiler(Mapping[str, Profile.Interval]):
    """Profile and record various events.

    !!! note

        * See [`Memory`][amltk.profiling.Memory] for more information on memory.
        * See [`Timer`][amltk.profiling.Timer] for more information on timing.

    Args:
        memory_unit: The default unit of memory to use.
        time_kind: The default type of timer to use.
    """

    profiles: dict[str, Profile.Interval] = field(default_factory=dict)
    memory_unit: Memory.Unit | Literal["B", "KB", "MB", "GB"] = "B"
    time_kind: Timer.Kind | Literal["wall", "cpu", "process"] = "wall"
    disabled: bool = False
    _running: deque[str] = field(default_factory=deque)

    @override
    def __getitem__(self, key: str) -> Profile.Interval:
        """Get a profile interval."""
        return self.profiles[key]

    @override
    def __iter__(self) -> Iterator[str]:
        """Iterate over the profile names."""
        return iter(self.profiles)

    @override
    def __len__(self) -> int:
        """Get the number of profiles."""
        return len(self.profiles)

    def disable(self) -> None:
        """Disable the profiler."""
        self.disabled = True

    def enable(self) -> None:
        """Enable the profiler."""
        self.disabled = False

    def each(
        self,
        itr: Iterable[T],
        *,
        name: str,
        itr_name: Callable[[int, T], str] | None = None,
    ) -> Iterator[T]:
        """Profile each item in an iterable.

        Args:
            itr: The iterable to profile.
            name: The name of the profile that lasts until iteration is complete
            itr_name: The name of the profile for each iteration.
                If a function is provided, it will be called with each item's index
                and the item. It should return a string. If `None` is provided,
                just the index will be used.

        Yields:
            The the items
        """
        if itr_name is None:
            itr_name = lambda i, _: str(i)
        with self.measure(name=name):
            for i, item in enumerate(itr):
                with self.measure(name=itr_name(i, item)):
                    yield item

    @contextmanager
    def __call__(
        self,
        name: str,
        *,
        memory_unit: Memory.Unit | Literal["B", "KB", "MB", "GB"] | None = None,
        time_kind: Timer.Kind | Literal["wall", "cpu", "process"] | None = None,
    ) -> Iterator[None]:
        """::: amltk.profiling.Profiler.measure"""  # noqa: D415
        with self.measure(name, memory_unit=memory_unit, time_kind=time_kind):
            yield

    @contextmanager
    def measure(
        self,
        name: str,
        *,
        memory_unit: Memory.Unit | Literal["B", "KB", "MB", "GB"] | None = None,
        time_kind: Timer.Kind | Literal["wall", "cpu", "process"] | None = None,
    ) -> Iterator[None]:
        """Profile a block of code. Store the result on this object.

        !!! note

            * See [`Memory`][amltk.profiling.Memory] for more information on memory.
            * See [`Timer`][amltk.profiling.Timer] for more information on timing.

        Args:
            name: The name of the profile.
            memory_unit: The unit of memory to use. Overwrites the default.
            time_kind: The type of timer to use. Overwrites the default.
        """
        if self.disabled:
            yield
            return

        memory_unit = memory_unit or self.memory_unit
        time_kind = time_kind or self.time_kind

        self._running.append(name)
        entry_name = ":".join(self._running)

        with Profile.measure(memory_unit=memory_unit, time_kind=time_kind) as profile:
            self.profiles[entry_name] = profile
            yield

        self._running.pop()

    def df(self) -> pd.DataFrame:
        """Convert the profiler to a dataframe."""
        return pd.DataFrame.from_dict(
            {k: v.to_dict() for k, v in self.profiles.items()},
            orient="index",
        )

    def __rich__(self) -> RenderableType:
        """Render the profiler."""
        from amltk._richutil import df_to_table

        _df = self.df()
        return df_to_table(_df, title="Profiler", index_style="bold")



"""Utilities for voting ensembles."""
from __future__ import annotations

from collections.abc import Iterable
from typing import TYPE_CHECKING, Any, TypeVar

import numpy as np
from sklearn.base import ClassifierMixin
from sklearn.ensemble import VotingClassifier, VotingRegressor
from sklearn.preprocessing import LabelEncoder, MultiLabelBinarizer
from sklearn.utils import Bunch

if TYPE_CHECKING:
    from sklearn.base import BaseEstimator

_Voter = TypeVar("_Voter", VotingRegressor, VotingClassifier)


def voting_with_preffited_estimators(
    estimators: Iterable[BaseEstimator],
    weights: Iterable[float] | None = None,
    *,
    voter: type[_Voter] | None = None,
    **voting_kwargs: Any,
) -> _Voter:
    """Create a voting ensemble with pre-fitted estimators.

    Args:
        estimators: The estimators to use in the ensemble.
        weights: The weights to use for the estimators. If None,
            will use uniform weights.
        voter: The voting classifier or regressor to use.
            If None, will use the appropriate one based on the type of the first
            estimator.
        **voting_kwargs: Additional arguments to pass to the voting classifier or
            regressor.

    Returns:
        The voting classifier or regressor with the pre-fitted estimators.
    """
    estimators = list(estimators)
    est0 = estimators[0]
    is_classification = voter is not None and issubclass(voter, VotingClassifier)

    if voter is None:
        if isinstance(est0, ClassifierMixin):
            voter_cls = VotingClassifier
            is_classification = True
        elif isinstance(est0, ClassifierMixin):
            voter_cls = VotingRegressor
            is_classification = False
        else:
            raise ValueError(
                f"Could not infer voter type from estimator type: {type(est0)}."
                " Please specify the voter type explicitly.",
            )
    else:
        voter_cls = voter

    if weights is None:
        weights = np.ones(len(estimators)) / len(estimators)
    else:
        weights = list(weights)

    named_estimators = [(str(i), e) for i, e in enumerate(estimators)]
    _voter = voter_cls(named_estimators, weights=weights, **voting_kwargs)
    _voter.estimators_ = [model for _, model in _voter.estimators]  # type: ignore

    if is_classification:
        est0_classes_ = est0.classes_  # type: ignore
        _voter.classes_ = est0_classes_  # type: ignore
        if np.ndim(est0_classes_) > 1:
            est0_classes_ = est0_classes_[0]
            _voter.le_ = MultiLabelBinarizer().fit(est0_classes_)  # type: ignore
        else:
            _voter.le_ = LabelEncoder().fit(est0.classes_)  # type: ignore

    _voter.named_estimators_ = Bunch()  # type: ignore

    # Taken from Sklearn _BaseVoting.fit
    # Uses 'drop' as placeholder for dropped estimators
    est_iter = iter(_voter.estimators_)  # type: ignore
    for name, est in _voter.estimators:  # type: ignore
        current_est = est if est == "drop" else next(est_iter)
        _voter.named_estimators_[name] = current_est  # type: ignore

        if hasattr(current_est, "feature_names_in_"):
            _voter.feature_names_in_ = current_est.feature_names_in_  # type: ignore

    return _voter  # type: ignore



r"""A plugin that can be attached to a Task.

By inheriting from a `Plugin`, you can hook into a
[`Task`][amltk.scheduling.Task]. A plugin can affect, modify and extend its
behaviours. Please see the documentation of the methods for more information.
Creating a plugin is only necesary if you need to modify actual behaviour of
the task. For siply hooking into the lifecycle of a task, you can use the `@events`
that a `Task` emits.

??? example "Creating a Plugin"

    For a full example of a simple plugin, see the
    [`Limiter`][amltk.scheduling.plugins.Limiter] plugin which prevents
    the task being submitted if for example, it has already been submitted
    too many times.

    The below example shows how to create a plugin that prints the task name
    before submitting it. It also emits an event when the task is submitted.

    ```python exec="true" source="material-block" html="true"
    from __future__ import annotations
    from typing import Callable

    from amltk.scheduling import Scheduler
    from amltk.scheduling.plugins import Plugin
    from amltk.scheduling.events import Event

    # A simple plugin that prints the task name before submitting
    class Printer(Plugin):
        name = "my-plugin"

        # Define an event the plugin will emit
        # Event[Task] indicates the callback for the event will be called with the task
        PRINTED: Event[str] = Event("printer-msg")

        def __init__(self, greeting: str):
            self.greeting = greeting
            self.n_greetings = 0

        def attach_task(self, task) -> None:
            self.task = task
            # Register an event with the task, this lets the task know valid events
            # people can subscribe to and helps it show up in visuals
            task.emitter.add_event(self.PRINTED)
            task.on_submitted(self._print_submitted, hidden=True)  # You can hide this callback from visuals

        def pre_submit(self, fn, *args, **kwargs) -> tuple[Callable, tuple, dict]:
            print(f"{self.greeting} for {self.task} {args} {kwargs}")
            self.n_greetings += 1
            return fn, args, kwargs

        def _print_submitted(self, future, *args, **kwargs) -> None:
            msg = f"Task was submitted {self.task} {args} {kwargs}"
            self.task.emitter.emit(self.PRINTED, msg)  # Emit the event with a msg

        def copy(self) -> Printer:
            # Plugins need to be able to copy themselves as if fresh
            return self.__class__(self.greeting)

        def __rich__(self):
            # Custome how the plugin is displayed in rich (Optional)
            # rich is an optional dependancy of amltk so we move the imports into here
            from rich.panel import Panel

            return Panel(
                f"Greeting: {self.greeting} ({self.n_greetings})",
                title=f"Plugin {self.name}"
            )

    def fn(x: int) -> int:
        return x + 1
    from amltk._doc import make_picklable; make_picklable(fn)  # markdown-exec: hide

    scheduler = Scheduler.with_processes(1)
    task = scheduler.task(fn, plugins=[Printer("Hello")])

    @scheduler.on_start
    def on_start():
        task.submit(15)

    @task.on("printer-msg")
    def callback(msg: str):
        print("\nmsg")

    scheduler.run()
    from amltk._doc import doc_print; doc_print(print, task)  # markdown-exec: hide
    ```

    All methods are optional, and you can choose to implement only the ones
    you need. Most plugins will likely need to implement the
    [`attach_task()`][amltk.scheduling.Plugin.attach_task] method, which is called
    when the plugin is attached to a task. In this method, you can for
    example subscribe to events on the task, create new subscribers for people
    to use or even store a reference to the task for later use.

    Plugins are also encouraged to utilize the events of a
    [`Task`][amltk.scheduling.Task] to further hook into the lifecycle of the task.
    For exampe, by saving a reference to the task in the `attach_task()` method, you
    can use the [`emit()`][amltk.scheduling.Task] method of the task to emit
    your own specialized events.
"""  # noqa: E501
from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from collections.abc import Callable
from itertools import chain
from typing import TYPE_CHECKING, ClassVar, TypeVar
from typing_extensions import ParamSpec, Self, override

from amltk._richutil.renderable import RichRenderable
from amltk.scheduling.events import Event

if TYPE_CHECKING:
    from rich.panel import Panel

    from amltk.scheduling import Task

logger = logging.getLogger(__name__)


P = ParamSpec("P")
P2 = ParamSpec("P2")

R = TypeVar("R")
R2 = TypeVar("R2")
CallableT = TypeVar("CallableT", bound=Callable)


class Plugin(RichRenderable, ABC):
    """A plugin that can be attached to a Task."""

    name: ClassVar[str]
    """The name of the plugin.

    This is used to identify the plugin during logging.
    """

    def attach_task(self, task: Task) -> None:
        """Attach the plugin to a task.

        This method is called when the plugin is attached to a task. This
        is the place to subscribe to events on the task, create new subscribers
        for people to use or even store a reference to the task for later use.

        Args:
            task: The task the plugin is being attached to.
        """

    def pre_submit(
        self,
        fn: Callable[P, R],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> tuple[Callable[P, R], tuple, dict] | None:
        """Pre-submit hook.

        This method is called before the task is submitted.

        Args:
            fn: The task function.
            *args: The arguments to the task function.
            **kwargs: The keyword arguments to the task function.

        Returns:
            A tuple of the task function, arguments and keyword arguments
            if the task should be submitted, or `None` if the task should
            not be submitted.
        """
        return fn, args, kwargs

    def events(self) -> list[Event]:
        """Return a list of events that this plugin emits.

        Likely no need to override this method, as it will automatically
        return all events defined on the plugin.
        """
        inherited_attrs = chain.from_iterable(
            vars(cls).values() for cls in self.__class__.__mro__
        )
        return [attr for attr in inherited_attrs if isinstance(attr, Event)]

    @abstractmethod
    def copy(self) -> Self:
        """Return a copy of the plugin.

        This method is used to create a copy of the plugin when a task is
        copied. This is useful if the plugin stores a reference to the task
        it is attached to, as the copy will need to store a reference to the
        copy of the task.
        """
        ...

    @override
    def __rich__(self) -> Panel:
        from rich.panel import Panel

        return Panel("", title=f"Plugin {self.name}")



"""The [`Scheduler`][amltk.scheduling.Scheduler] uses
an [`Executor`][concurrent.futures.Executor], a builtin python native with
a `#!python submit(f, *args, **kwargs)` function to submit compute to
be compute else where, whether it be locally or remotely.

The `Scheduler` is primarily used to dispatch compute to an `Executor` and
emit `@events`, which can trigger user callbacks.

Typically you should not use the `Scheduler` directly for dispatching and
responding to computed functions, but rather use a [`Task`][amltk.scheduling.Task]

??? note "Running in a Jupyter Notebook/Colab"

    If you are using a Jupyter Notebook, you likley need to use the following
    at the top of your notebook:

    ```python
    import nest_asyncio  # Only necessary in Notebooks
    nest_asyncio.apply()

    scheduler.run(...)
    ```

    This is due to the fact a notebook runs in an async context. If you do not
    wish to use the above snippet, you can instead use:

    ```python
    await scheduler.async_run(...)
    ```

??? tip "Basic Usage"

    In this example, we create a scheduler that uses local processes as
    workers. We then create a task that will run a function `fn` and submit it
    to the scheduler. Lastly, a callback is registered to `@on_future_result` to print the
    result when the compute is done.

    ```python exec="true" source="material-block" html="true"
    from amltk.scheduling import Scheduler

    def fn(x: int) -> int:
        return x + 1
    from amltk._doc import make_picklable; make_picklable(fn)  # markdown-exec: hide

    scheduler = Scheduler.with_processes(1)

    @scheduler.on_start
    def launch_the_compute():
        scheduler.submit(fn, 1)

    @scheduler.on_future_result
    def callback(future, result):
        print(f"Result: {result}")

    scheduler.run()
    from amltk._doc import doc_print; doc_print(print, scheduler)  # markdown-exec: hide
    ```

    The last line in the previous example called
    [`scheduler.run()`][amltk.scheduling.Scheduler.run] is what starts the scheduler
    running, in which it will first emit the `@on_start` event. This triggered the
    callback `launch_the_compute()` which submitted the function `fn` with the
    arguments `#!python 1`.

    The scheduler then ran the compute and waited for it to complete, emitting the
    `@on_future_result` event when it was done successfully. This triggered the callback
    `callback()` which printed the result.

    At this point, there is no more compute happening and no more events to respond to
    so the scheduler will halt.

??? example "`@events`"

    === "Scheduler Status Events"

        When the scheduler enters some important state, it will emit an event
        to let you know.

        === "`@on_start`"

            ::: amltk.scheduling.Scheduler.on_start

        === "`@on_finishing`"

            ::: amltk.scheduling.Scheduler.on_finishing

        === "`@on_finished`"

            ::: amltk.scheduling.Scheduler.on_finished

        === "`@on_stop`"

            ::: amltk.scheduling.Scheduler.on_stop

        === "`@on_timeout`"

            ::: amltk.scheduling.Scheduler.on_timeout

        === "`@on_empty`"

            ::: amltk.scheduling.Scheduler.on_empty

    === "Submitted Compute Events"

        When any compute goes through the `Scheduler`, it will emit an event
        to let you know. You should however prefer to use a
        [`Task`][amltk.scheduling.Task] as it will emit specific events
        for the task at hand, and not all compute.

        === "`@on_future_submitted`"

            ::: amltk.scheduling.Scheduler.on_future_submitted

        === "`@on_future_result`"

            ::: amltk.scheduling.Scheduler.on_future_result

        === "`@on_future_exception`"

            ::: amltk.scheduling.Scheduler.on_future_exception

        === "`@on_future_done`"

            ::: amltk.scheduling.Scheduler.on_future_done

        === "`@on_future_cancelled`"

            ::: amltk.scheduling.Scheduler.on_future_cancelled


??? tip "Common usages of `run()`"

    There are various ways to [`run()`][amltk.scheduling.Scheduler.run] the
    scheduler, notably how long it should run with `timeout=` and also how
    it should react to any exception that may have occurred within the `Scheduler`
    itself or your callbacks.

    Please see the [`run()`][amltk.scheduling.Scheduler.run] API doc for more
    details and features, however we show two common use cases of using the `timeout=`
    parameter.

    You can render a live display using [`run(display=...)`][amltk.scheduling.Scheduler.run].
    This require [`rich`](https://github.com/Textualize/rich) to be installed. You
    can install this with `#!bash pip install rich` or `#!bash pip install amltk[rich]`.


    === "`run(timeout=...)`"

        You can tell the `Scheduler` to stop after a certain amount of time
        with the `timeout=` argument to [`run()`][amltk.scheduling.Scheduler.run].

        This will also trigger the `@on_timeout` event as seen in the `Scheduler` output.

        ```python exec="true" source="material-block" html="True" hl_lines="19"
        import time
        from amltk.scheduling import Scheduler

        scheduler = Scheduler.with_processes(1)

        def expensive_function() -> int:
            time.sleep(0.1)
            return 42
        from amltk._doc import make_picklable; make_picklable(expensive_function)  # markdown-exec: hide

        @scheduler.on_start
        def submit_calculations() -> None:
            scheduler.submit(expensive_function)

        # The will endlessly loop the scheduler
        @scheduler.on_future_done
        def submit_again(future: Future) -> None:
            if scheduler.running():
                scheduler.submit(expensive_function)

        scheduler.run(timeout=1)  # End after 1 second
        from amltk._doc import doc_print; doc_print(print, scheduler, output="html", fontsize="small")  # markdown-exec: hide
        ```

    === "`run(timeout=..., wait=False)`"

        By specifying that the `Scheduler` should not wait for ongoing tasks
        to finish, the `Scheduler` will attempt to cancel and possibly terminate
        any running tasks.

        ```python exec="true" source="material-block" html="True"
        import time
        from amltk.scheduling import Scheduler

        scheduler = Scheduler.with_processes(1)

        def expensive_function() -> None:
            time.sleep(10)

        from amltk._doc import make_picklable; make_picklable(expensive_function)  # markdown-exec: hide

        @scheduler.on_start
        def submit_calculations() -> None:
            scheduler.submit(expensive_function)

        scheduler.run(timeout=1, wait=False)  # End after 1 second
        from amltk._doc import doc_print; doc_print(print, scheduler, output="html", fontsize="small")  # markdown-exec: hide
        ```

        ??? info "Forcibly Terminating Workers"

            As an `Executor` does not provide an interface to forcibly
            terminate workers, we provide `Scheduler(terminate=...)` as a custom
            strategy for cleaning up a provided executor. It is not possible
            to terminate running thread based workers, for example using
            `ThreadPoolExecutor` and any Executor using threads to spawn
            tasks will have to wait until all running tasks are finish
            before python can close.

            It's likely `terminate` will trigger the `EXCEPTION` event for
            any tasks that are running during the shutdown, **not***
            a cancelled event. This is because we use a
            [`Future`][concurrent.futures.Future]
            under the hood and these can not be cancelled once running.
            However there is no guarantee of this and is up to how the
            `Executor` handles this.

??? example "Scheduling something to be run later"

    You can schedule some function to be run later using the
    [`#!python scheduler.call_later()`][amltk.scheduling.Scheduler.call_later] method.

    !!! note

        This does not run the function in the background, it just schedules some
        function to be called later, where you could perhaps then use submit to
        scheduler a [`Task`][amltk.scheduling.Task] to run the function in the
        background.

    ```python exec="true" source="material-block" result="python"
    from amltk.scheduling import Scheduler

    scheduler = Scheduler.with_processes(1)

    def fn() -> int:
        print("Ending now!")
        scheduler.stop()

    @scheduler.on_start
    def schedule_fn() -> None:
        scheduler.call_later(1, fn)

    scheduler.run(end_on_empty=False)
    ```

"""  # noqa: E501
from __future__ import annotations

import asyncio
import logging
import warnings
from asyncio import Future
from collections.abc import Callable, Iterable, Mapping
from concurrent.futures import CancelledError, Executor, ProcessPoolExecutor
from contextlib import nullcontext
from dataclasses import dataclass
from enum import Enum, auto
from functools import partial
from threading import Timer
from typing import (
    TYPE_CHECKING,
    Any,
    Concatenate,
    Literal,
    ParamSpec,
    TypeVar,
    overload,
)
from typing_extensions import Self, override
from uuid import uuid4

from amltk._asyncm import ContextEvent
from amltk._functional import Flag, subclass_map
from amltk._richutil.renderable import RichRenderable
from amltk._util import ignore_warnings, mutli_context
from amltk.exceptions import SchedulerNotRunningError
from amltk.scheduling.events import Emitter, Event, Subscriber
from amltk.scheduling.executors import SequentialExecutor
from amltk.scheduling.task import Task
from amltk.scheduling.termination_strategies import termination_strategy

if TYPE_CHECKING:
    from multiprocessing.context import BaseContext

    from rich.console import RenderableType
    from rich.live import Live

    from amltk.scheduling.executors.dask_jobqueue import DJQ_NAMES
    from amltk.scheduling.plugins import Plugin
    from amltk.scheduling.plugins.comm import Comm

    P = ParamSpec("P")
    R = TypeVar("R")


logger = logging.getLogger(__name__)


class Scheduler(RichRenderable):
    """A scheduler for submitting tasks to an Executor."""

    executor: Executor
    """The executor to use to run tasks."""

    emitter: Emitter
    """The emitter to use for events."""

    queue: dict[Future, tuple[Callable, tuple, dict]]
    """The queue of tasks running."""

    on_start: Subscriber[[]]
    """A [`Subscriber`][amltk.scheduling.events.Subscriber] which is called when the
    scheduler starts. This is the first event emitted by the scheduler and
    one of the only ways to submit the initial compute to the scheduler.

    ```python
    @scheduler.on_start
    def my_callback():
        ...
    ```
    """
    on_future_submitted: Subscriber[Future]
    """A [`Subscriber`][amltk.scheduling.events.Subscriber] which is called when
    some compute is submitted.

    ```python
    @scheduler.on_future_submitted
    def my_callback(future: Future):
        ...
    ```
    """
    on_future_done: Subscriber[Future]
    """A [`Subscriber`][amltk.scheduling.events.Subscriber] which is called when
    some compute is done, regardless of whether it was successful or not.

    ```python
    @scheduler.on_future_done
    def my_callback(future: Future):
        ...
    ```
    """
    on_future_result: Subscriber[Future, Any]
    """A [`Subscriber`][amltk.scheduling.events.Subscriber] which is called when
    a future returned with a result, no exception raise.

    ```python
    @scheduler.on_future_result
    def my_callback(future: Future, result: Any):
        ...
    ```
    """
    on_future_exception: Subscriber[Future, BaseException]
    """A [`Subscriber`][amltk.scheduling.events.Subscriber] which is called when
    some compute raised an uncaught exception.

    ```python
    @scheduler.on_future_exception
    def my_callback(future: Future, exception: BaseException):
        ...
    ```
    """
    on_future_cancelled: Subscriber[Future]
    """A [`Subscriber`][amltk.scheduling.events.Subscriber] which is called
    when a future is cancelled. This usually occurs due to the underlying Scheduler,
    and is not something we do directly, other than when shutting down the scheduler.

    ```python
    @scheduler.on_future_cancelled
    def my_callback(future: Future):
        ...
    ```
    """
    on_finishing: Subscriber[[]]
    """A [`Subscriber`][amltk.scheduling.events.Subscriber] which is called when the
    scheduler is finishing up. This occurs right before the scheduler shuts down
    the executor.

    ```python
    @scheduler.on_finishing
    def my_callback():
        ...
    ```
    """
    on_finished: Subscriber[[]]
    """A [`Subscriber`][amltk.scheduling.events.Subscriber] which is called when
    the scheduler is finished, has shutdown the executor and possibly
    terminated any remaining compute.

    ```python
    @scheduler.on_finished
    def my_callback():
        ...
    ```
    """
    on_stop: Subscriber[str, BaseException | None]
    """A [`Subscriber`][amltk.scheduling.events.Subscriber] which is called when the
    scheduler is has been stopped due to the [`stop()`][amltk.scheduling.Scheduler.stop]
    method being called.

    ```python
    @scheduler.on_stop
    def my_callback(stop_msg: str, exception: BaseException | None):
        ...
    ```
    """
    on_timeout: Subscriber[[]]
    """A [`Subscriber`][amltk.scheduling.events.Subscriber] which is called when
    the scheduler reaches the timeout.

    ```python
    @scheduler.on_timeout
    def my_callback():
        ...
    ```
    """
    on_empty: Subscriber[[]]
    """A [`Subscriber`][amltk.scheduling.events.Subscriber] which is called when the
    queue is empty. This can be useful to re-fill the queue and prevent the
    scheduler from exiting.

    ```python
    @scheduler.on_empty
    def my_callback():
        ...
    ```
    """

    STARTED: Event[[]] = Event("on_start")
    FINISHING: Event[[]] = Event("on_finishing")
    FINISHED: Event[[]] = Event("on_finished")
    STOP: Event[str, BaseException | None] = Event("on_stop")
    TIMEOUT: Event[[]] = Event("on_timeout")
    EMPTY: Event[[]] = Event("on_empty")
    FUTURE_SUBMITTED: Event[Future] = Event("on_future_submitted")
    FUTURE_DONE: Event[Future] = Event("on_future_done")
    FUTURE_CANCELLED: Event[Future] = Event("on_future_cancelled")
    FUTURE_RESULT: Event[Future, Any] = Event("on_future_result")
    FUTURE_EXCEPTION: Event[Future, BaseException] = Event("on_future_exception")

    def __init__(
        self,
        executor: Executor,
        *,
        terminate: Callable[[Executor], None] | bool = True,
    ) -> None:
        """Initialize a scheduler.

        Args:
            executor: The dispatcher to use for submitting tasks.
            terminate: Whether to call shutdown on the executor when
                `run(..., wait=False)`. If True, the executor will be
                `shutdown(wait=False)` and we will attempt to terminate
                any workers of the executor. For some `Executors` this
                is enough, i.e. Dask, however for something like
                `ProcessPoolExecutor`, we will use `psutil` to kill
                its worker processes. If a callable, we will use this
                function for custom worker termination.
                If False, shutdown will not be called and the executor will
                remain active.
        """
        super().__init__()
        self.executor = executor
        self.unique_ref = f"Scheduler-{uuid4()}"
        self.emitter = Emitter()
        self.event_counts = self.emitter.event_counts

        # The current state of things and references to them
        self.queue = {}

        # Set up subscribers for events
        self.on_start = self.emitter.subscriber(self.STARTED)
        self.on_finishing = self.emitter.subscriber(self.FINISHING)
        self.on_finished = self.emitter.subscriber(self.FINISHED)
        self.on_stop = self.emitter.subscriber(self.STOP)
        self.on_timeout = self.emitter.subscriber(self.TIMEOUT)
        self.on_empty = self.emitter.subscriber(self.EMPTY)

        self.on_future_submitted = self.emitter.subscriber(self.FUTURE_SUBMITTED)
        self.on_future_done = self.emitter.subscriber(self.FUTURE_DONE)
        self.on_future_cancelled = self.emitter.subscriber(self.FUTURE_CANCELLED)
        self.on_future_exception = self.emitter.subscriber(self.FUTURE_EXCEPTION)
        self.on_future_result = self.emitter.subscriber(self.FUTURE_RESULT)

        self._terminate: Callable[[Executor], None] | None
        if terminate is True:
            self._terminate = termination_strategy(executor)
        else:
            self._terminate = terminate if callable(terminate) else None

        # This can be triggered either by `scheduler.stop` in a callback.
        # Has to be created inside the event loop so there's no issues
        self._stop_event: ContextEvent | None = None

        # This is a condition to make sure monitoring the queue will wait properly
        self._queue_has_items_event = asyncio.Event()

        # This is triggered when run is called
        self._running_event = ContextEvent()

        # This is set once `run` is called.
        # Either contains the mapping from exception to what to do,
        # otherwise a boolean whether to end or not.
        self._on_exc_method_map: Flag[
            dict[type[Exception], Literal["raise", "end", "continue"]],
        ] = Flag(initial={Exception: "raise"})

        # This is set once `run` is called.
        # Indicates what to do when a cancellation occurs
        self._on_cancel_method: Flag[Literal["raise", "end", "continue"]] = Flag(
            initial="raise",
        )

        # This is used to manage suequential queues, where we need a Thread
        # timer to ensure that we don't get caught in an endless loop waiting
        # for the `timeout` in `_run_scheduler` to trigger. This won't trigger
        # because the sync code of submit could possibly keep calling itself
        # endlessly, preventing any of the async code from running.
        self._timeout_timer: Timer | None = None

        # A collection of things that want to register as being part of something
        # to render when the Scheduler is rendered.
        self._renderables: list[RenderableType] = [self.emitter]

        # These are extra user provided renderables during a call to `run()`. We
        # seperate these out so that we can remove them when the scheduler is
        # stopped.
        self._extra_renderables: list[RenderableType] | None = None

        # An indicator an object to render live output (if requested) with
        # `display=` on a call to `run()`
        self._live_output: Live | None = None

    @classmethod
    def with_processes(
        cls,
        max_workers: int | None = None,
        mp_context: BaseContext | Literal["fork", "spawn", "forkserver"] | None = None,
        initializer: Callable[..., Any] | None = None,
        initargs: tuple[Any, ...] = (),
    ) -> Self:
        """Create a scheduler with a `ProcessPoolExecutor`.

        See [`ProcessPoolExecutor`][concurrent.futures.ProcessPoolExecutor]
        for more details.
        """
        if isinstance(mp_context, str):
            from multiprocessing import get_context

            mp_context = get_context(mp_context)

        executor = ProcessPoolExecutor(
            max_workers=max_workers,
            mp_context=mp_context,
            initializer=initializer,
            initargs=initargs,
        )
        return cls(executor=executor)

    @classmethod
    def with_loky(  # noqa: PLR0913
        cls,
        max_workers: int | None = None,
        context: BaseContext | Literal["fork", "spawn", "forkserver"] | None = None,
        timeout: int = 10,
        kill_workers: bool = False,  # noqa: FBT002, FBT001
        reuse: bool | Literal["auto"] = "auto",
        job_reducers: Any | None = None,
        result_reducers: Any | None = None,
        initializer: Callable[..., Any] | None = None,
        initargs: tuple[Any, ...] = (),
        env: dict[str, str] | None = None,
    ) -> Self:
        """Create a scheduler with a `loky.get_reusable_executor`.

        See [loky documentation][https://loky.readthedocs.io/en/stable/API.html]
        for more details.
        """
        from loky import get_reusable_executor

        executor = get_reusable_executor(
            max_workers=max_workers,
            context=context,
            timeout=timeout,
            kill_workers=kill_workers,
            reuse=reuse,  # type: ignore
            job_reducers=job_reducers,
            result_reducers=result_reducers,
            initializer=initializer,
            initargs=initargs,
            env=env,
        )
        return cls(executor=executor)

    @classmethod
    def with_sequential(cls) -> Self:
        """Create a Scheduler that runs sequentially.

        This is useful for debugging and testing. Uses
        a [`SequentialExecutor`][amltk.scheduling.SequentialExecutor].
        """
        return cls(executor=SequentialExecutor())

    @classmethod
    def with_slurm(
        cls,
        *,
        n_workers: int,
        adaptive: bool = False,
        submit_command: str | None = None,
        cancel_command: str | None = None,
        **kwargs: Any,
    ) -> Self:
        """Create a Scheduler that runs on a SLURM cluster.

        This is useful for running on a SLURM cluster. Uses
        [dask_jobqueue.SLURMCluster][].

        Args:
            n_workers: The number of workers to start.
            adaptive: Whether to use the adaptive scaling of the cluster or fixed
                allocate all workers. This will specifically use the
                [dask_jobqueue.SLURMCluster.adapt](https://jobqueue.dask.org/en/latest/index.html?highlight=adapt#adaptivity)
                method to dynamically scale the cluster to the number of workers
                specified.
            submit_command: Overwrite the command to submit a worker if necessary.
            cancel_command: Overwrite the command to cancel a worker if necessary.
            kwargs: Any additional keyword arguments to pass to the
                `dask_jobqueue` class.

        Returns:
            A scheduler that will run on a SLURM cluster.
        """
        return cls.with_dask_jobqueue(
            "slurm",
            n_workers=n_workers,
            adaptive=adaptive,
            submit_command=submit_command,
            cancel_command=cancel_command,
            **kwargs,
        )

    @classmethod
    def with_pbs(
        cls,
        *,
        n_workers: int,
        adaptive: bool = False,
        submit_command: str | None = None,
        cancel_command: str | None = None,
        **kwargs: Any,
    ) -> Self:
        """Create a Scheduler that runs on a PBS cluster.

        This is useful for running on a PBS cluster. Uses
        [dask_jobqueue.PBSCluster][].

        Args:
            n_workers: The number of workers to start.
            adaptive: Whether to use the adaptive scaling of the cluster or fixed
                allocate all workers. This will specifically use the
                [dask_jobqueue.SLURMCluster.adapt](https://jobqueue.dask.org/en/latest/index.html?highlight=adapt#adaptivity)
                method to dynamically scale the cluster to the number of workers
                specified.
            submit_command: Overwrite the command to submit a worker if necessary.
            cancel_command: Overwrite the command to cancel a worker if necessary.
            kwargs: Any additional keyword arguments to pass to the
                `dask_jobqueue` class.

        Returns:
            A scheduler that will run on a PBS cluster.
        """
        return cls.with_dask_jobqueue(
            "pbs",
            n_workers=n_workers,
            adaptive=adaptive,
            submit_command=submit_command,
            cancel_command=cancel_command,
            **kwargs,
        )

    @classmethod
    def with_sge(
        cls,
        *,
        n_workers: int,
        adaptive: bool = False,
        submit_command: str | None = None,
        cancel_command: str | None = None,
        **kwargs: Any,
    ) -> Self:
        """Create a Scheduler that runs on a SGE cluster.

        This is useful for running on a SGE cluster. Uses
        [dask_jobqueue.SGECluster][].

        Args:
            n_workers: The number of workers to start.
            adaptive: Whether to use the adaptive scaling of the cluster or fixed
                allocate all workers. This will specifically use the
                [dask_jobqueue.SLURMCluster.adapt](https://jobqueue.dask.org/en/latest/index.html?highlight=adapt#adaptivity)
                method to dynamically scale the cluster to the number of workers
                specified.
            submit_command: Overwrite the command to submit a worker if necessary.
            cancel_command: Overwrite the command to cancel a worker if necessary.
            kwargs: Any additional keyword arguments to pass to the
                `dask_jobqueue` class.

        Returns:
            A scheduler that will run on a SGE cluster.
        """
        return cls.with_dask_jobqueue(
            "sge",
            n_workers=n_workers,
            adaptive=adaptive,
            submit_command=submit_command,
            cancel_command=cancel_command,
            **kwargs,
        )

    @classmethod
    def with_oar(
        cls,
        *,
        n_workers: int,
        adaptive: bool = False,
        submit_command: str | None = None,
        cancel_command: str | None = None,
        **kwargs: Any,
    ) -> Self:
        """Create a Scheduler that runs on a OAR cluster.

        This is useful for running on a OAR cluster. Uses
        [dask_jobqueue.OARCluster][].

        Args:
            n_workers: The number of workers to start.
            adaptive: Whether to use the adaptive scaling of the cluster or fixed
                allocate all workers. This will specifically use the
                [dask_jobqueue.SLURMCluster.adapt](https://jobqueue.dask.org/en/latest/index.html?highlight=adapt#adaptivity)
                method to dynamically scale the cluster to the number of workers
                specified.
            submit_command: Overwrite the command to submit a worker if necessary.
            cancel_command: Overwrite the command to cancel a worker if necessary.
            kwargs: Any additional keyword arguments to pass to the
                `dask_jobqueue` class.

        Returns:
            A scheduler that will run on a OAR cluster.
        """
        return cls.with_dask_jobqueue(
            "oar",
            n_workers=n_workers,
            adaptive=adaptive,
            submit_command=submit_command,
            cancel_command=cancel_command,
            **kwargs,
        )

    @classmethod
    def with_moab(
        cls,
        *,
        n_workers: int,
        adaptive: bool = False,
        submit_command: str | None = None,
        cancel_command: str | None = None,
        **kwargs: Any,
    ) -> Self:
        """Create a Scheduler that runs on a Moab cluster.

        This is useful for running on a Moab cluster. Uses
        [dask_jobqueue.MoabCluster][].

        Args:
            n_workers: The number of workers to start.
            adaptive: Whether to use the adaptive scaling of the cluster or fixed
                allocate all workers. This will specifically use the
                [dask_jobqueue.SLURMCluster.adapt](https://jobqueue.dask.org/en/latest/index.html?highlight=adapt#adaptivity)
                method to dynamically scale the cluster to the number of workers
                specified.
            submit_command: Overwrite the command to submit a worker if necessary.
            cancel_command: Overwrite the command to cancel a worker if necessary.
            kwargs: Any additional keyword arguments to pass to the
                `dask_jobqueue` class.

        Returns:
            A scheduler that will run on a Moab cluster.
        """
        return cls.with_dask_jobqueue(
            "moab",
            n_workers=n_workers,
            adaptive=adaptive,
            submit_command=submit_command,
            cancel_command=cancel_command,
            **kwargs,
        )

    @classmethod
    def with_lsf(
        cls,
        *,
        n_workers: int,
        adaptive: bool = False,
        submit_command: str | None = None,
        cancel_command: str | None = None,
        **kwargs: Any,
    ) -> Self:
        """Create a Scheduler that runs on a LSF cluster.

        This is useful for running on a LSF cluster. Uses
        [dask_jobqueue.LSFCluster][].

        Args:
            n_workers: The number of workers to start.
            adaptive: Whether to use the adaptive scaling of the cluster or fixed
                allocate all workers. This will specifically use the
                [dask_jobqueue.SLURMCluster.adapt](https://jobqueue.dask.org/en/latest/index.html?highlight=adapt#adaptivity)
                method to dynamically scale the cluster to the number of workers
                specified.
            submit_command: Overwrite the command to submit a worker if necessary.
            cancel_command: Overwrite the command to cancel a worker if necessary.
            kwargs: Any additional keyword arguments to pass to the
                `dask_jobqueue` class.

        Returns:
            A scheduler that will run on a LSF cluster.
        """
        return cls.with_dask_jobqueue(
            "lsf",
            n_workers=n_workers,
            adaptive=adaptive,
            submit_command=submit_command,
            cancel_command=cancel_command,
            **kwargs,
        )

    @classmethod
    def with_htcondor(
        cls,
        *,
        n_workers: int,
        adaptive: bool = False,
        submit_command: str | None = None,
        cancel_command: str | None = None,
        **kwargs: Any,
    ) -> Self:
        """Create a Scheduler that runs on a HTCondor cluster.

        This is useful for running on a HTCondor cluster. Uses
        [dask_jobqueue.HTCondorCluster][].

        Args:
            n_workers: The number of workers to start.
            adaptive: Whether to use the adaptive scaling of the cluster or fixed
                allocate all workers. This will specifically use the
                [dask_jobqueue.SLURMCluster.adapt](https://jobqueue.dask.org/en/latest/index.html?highlight=adapt#adaptivity)
                method to dynamically scale the cluster to the number of workers
                specified.
            submit_command: Overwrite the command to submit a worker if necessary.
            cancel_command: Overwrite the command to cancel a worker if necessary.
            kwargs: Any additional keyword arguments to pass to the
                `dask_jobqueue` class.

        Returns:
            A scheduler that will run on a HTCondor cluster.
        """
        return cls.with_dask_jobqueue(
            "htcondor",
            n_workers=n_workers,
            adaptive=adaptive,
            submit_command=submit_command,
            cancel_command=cancel_command,
            **kwargs,
        )

    @classmethod
    def with_dask_jobqueue(
        cls,
        name: DJQ_NAMES,
        *,
        n_workers: int,
        adaptive: bool = False,
        submit_command: str | None = None,
        cancel_command: str | None = None,
        **kwargs: Any,
    ) -> Self:
        """Create a Scheduler with using `dask-jobqueue`.

        See [`dask_jobqueue`][dask_jobqueue] for more details.

        [dask_jobqueue]: https://jobqueue.dask.org/en/latest/

        Args:
            name: The name of the jobqueue to use. This is the name of the
                class in `dask_jobqueue` to use. For example, to use
                `dask_jobqueue.SLURMCluster`, you would use `slurm`.
            adaptive: Whether to use the adaptive scaling of the cluster or fixed
                allocate all workers. This will specifically use the
                [dask_jobqueue.SLURMCluster.adapt](https://jobqueue.dask.org/en/latest/index.html?highlight=adapt#adaptivity)
                method to dynamically scale the cluster to the number of workers
                specified.
            n_workers: The number of workers to start.
            submit_command: Overwrite the command to submit a worker if necessary.
            cancel_command: Overwrite the command to cancel a worker if necessary.
            kwargs: Any additional keyword arguments to pass to the
                `dask_jobqueue` class.

        Raises:
            ImportError: If `dask-jobqueue` is not installed.

        Returns:
            A new scheduler with a `dask_jobqueue` executor.
        """
        try:
            from amltk.scheduling.executors.dask_jobqueue import DaskJobqueueExecutor

        except ImportError as e:
            raise ImportError(
                f"To use the {name} executor, you must install the "
                "`dask-jobqueue` package.",
            ) from e

        executor = DaskJobqueueExecutor.from_str(
            name,
            n_workers=n_workers,
            adaptive=adaptive,
            submit_command=submit_command,
            cancel_command=cancel_command,
            **kwargs,
        )
        return cls(executor)

    def empty(self) -> bool:
        """Check if the scheduler is empty.

        Returns:
            True if there are no tasks in the queue.
        """
        return len(self.queue) == 0

    def running(self) -> bool:
        """Whether the scheduler is running and accepting tasks to dispatch.

        Returns:
            True if the scheduler is running and accepting tasks.
        """
        return self._running_event.is_set()

    def submit(
        self,
        fn: Callable[P, R],
        /,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> Future[R]:
        """Submits a callable to be executed with the given arguments.

        Args:
            fn: The callable to be executed as
                fn(*args, **kwargs) that returns a Future instance representing
                the execution of the callable.
            args: positional arguments to pass to the function
            kwargs: keyword arguments to pass to the function

        Raises:
            SchedulerNotRunningError: If the scheduler is not running.
                You can protect against this using,
                [`scheduler.running()`][amltk.scheduling.scheduler.Scheduler.running].

        Returns:
            A Future representing the given call.
        """
        if not self.running():
            msg = (
                f"Scheduler is not running, cannot submit task {fn}"
                f" with {args=}, {kwargs=}"
            )
            raise SchedulerNotRunningError(msg)

        try:
            sync_future = self.executor.submit(fn, *args, **kwargs)
            future = asyncio.wrap_future(sync_future)
        except Exception as e:
            logger.exception(f"Could not submit task {fn}", exc_info=e)
            raise e

        self._register_future(future, fn, *args, **kwargs)
        return future

    @overload
    def task(
        self,
        function: Callable[Concatenate[Comm, P], R],
        *,
        plugins: Comm.Plugin | Iterable[Comm.Plugin | Plugin] = ...,
        init_plugins: bool = ...,
    ) -> Task[P, R]:
        ...

    @overload
    def task(
        self,
        function: Callable[P, R],
        *,
        plugins: Plugin | Iterable[Plugin] = (),
        init_plugins: bool = True,
    ) -> Task[P, R]:
        ...

    def task(
        self,
        function: Callable[P, R] | Callable[Concatenate[Comm, P], R],
        *,
        plugins: Plugin | Iterable[Plugin] = (),
        init_plugins: bool = True,
    ) -> Task[P, R]:
        """Create a new task.

        Args:
            function: The function to run using the scheduler.
            plugins: The plugins to attach to the task.
            init_plugins: Whether to initialize the plugins.

        Returns:
            A new task.
        """
        # HACK: Not that the type: ignore is due to the fact that we can't use type
        # checking to enforce that
        # A. `function` is a callable with the first arg being a Comm
        # B. `plugins`
        task = Task(function, self, plugins=plugins, init_plugins=init_plugins)  # type: ignore
        self.add_renderable(task)
        return task  # type: ignore

    def _register_future(
        self,
        future: Future,
        function: Callable,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """Registers the future into the queue and add a callback that will be called
        upon future completion. This callback will remove the future from the queue.
        """
        self.queue[future] = (function, args, kwargs)
        self._queue_has_items_event.set()

        self.on_future_submitted.emit(future)
        # Display if requested
        if self._live_output:
            self._live_output.refresh()
            future.add_done_callback(
                lambda _, live=self._live_output: live.refresh(),  # type: ignore
            )

        future.add_done_callback(self._register_complete)

    def _register_complete(self, future: Future) -> None:
        try:
            self.queue.pop(future)
        except ValueError as e:
            logger.error(
                f"{future=} was not found in the queue {self.queue}: {e}!",
                exc_info=True,
            )

        if future.cancelled():
            self.on_future_cancelled.emit(future)
            match self._on_cancel_method.value:
                case "raise" | "end":
                    self.stop(
                        stop_msg="Ending on cancellation of a future",
                        exception=CancelledError(f"Future {future} was cancelled"),
                    )
                case "continue":
                    pass
            return

        self.on_future_done.emit(future)

        exception = future.exception()
        match exception:
            case None:
                result = future.result()
                self.on_future_result.emit(future, result)
                return
            # We do not allow users to handle `BaseException`s as they are
            # not meant to be caught.
            case BaseException() if not isinstance(exception, Exception):
                self.on_future_exception.emit(future, exception)
                self.stop(
                    stop_msg=(
                        f"BaseException occured - {type(exception)}:{exception}."
                        "\nEnding as this is not meant to be caught."
                    ),
                    exception=exception,
                )
            case Exception():
                self.on_future_exception.emit(future, exception)
                match subclass_map(exception, self._on_exc_method_map.value):
                    case (err_type, "continue"):
                        return
                    case (err_type, "end"):
                        self.stop(
                            stop_msg=(
                                f"Ending on exception '{type(exception)}'"
                                f" as {err_type} is 'end' as specified from"
                                f"`on_exception={self._on_exc_method_map.value}"
                            ),
                            exception=exception,
                        )
                        return
                    case (err_type, "raise"):
                        self.stop(
                            stop_msg=(
                                f"raising on exception '{type(exception)}'"
                                f" as {err_type} is 'raise' as specified from"
                                f"`on_exception={self._on_exc_method_map.value}"
                            ),
                            exception=exception,
                        )
                        return
                    case None:
                        self.stop(
                            stop_msg=(
                                f"Ending on exception type {type(exception)} as"
                                "it's not contained in"
                                f"`on_exception={self._on_exc_method_map.value}"
                            ),
                            exception=exception,
                        )
                        return
                    case _:
                        pass
            case _:
                pass

        raise RuntimeError(
            f"Unexpected thing from future: {exception=}\nPlease report this as a bug!",
        ) from exception

    async def _monitor_queue_empty(self) -> None:
        """Monitor for the queue being empty and trigger an event when it is."""
        if not self.running():
            raise RuntimeError("The scheduler is not running!")

        while True:
            while self.queue:
                queue = list(self.queue)
                await asyncio.wait(queue, return_when=asyncio.ALL_COMPLETED)

            # Signal that the queue is now empty
            self._queue_has_items_event.clear()
            self.on_empty.emit()

            # Wait for an item to be in the queue
            await self._queue_has_items_event.wait()

            logger.debug("Queue has been filled again")

    async def _stop_when_triggered(self, stop_event: ContextEvent) -> bool:
        """Stop the scheduler when the stop event is set."""
        if not self.running():
            raise RuntimeError("The scheduler is not running!")

        await stop_event.wait()

        logger.debug("Stop event triggered, stopping scheduler")
        return True

    async def _run_scheduler(  # noqa: C901, PLR0915
        self,
        *,
        timeout: float | None = None,
        end_on_empty: bool = True,
        wait: bool = True,
    ) -> ExitState.Code | BaseException:
        with mutli_context(
            self.executor,
            self._live_output if self._live_output is not None else nullcontext(),
            ignore_warnings() if self._live_output is not None else nullcontext(),
        ):
            self._stop_event = ContextEvent()

            # Start a Thread Timer as our timing mechanism.
            # HACK: This is required because the SequentialExecutor mode
            # will not allow the async loop to run, meaning we can't update
            # any internal state. For this reason, we use a Thread
            if timeout is not None:
                self._timeout_timer = Timer(timeout, lambda: None)
                self._timeout_timer.start()

            with self._running_event:
                self.on_start.emit()

                # Monitor for `stop` being triggered
                stop_triggered = asyncio.create_task(
                    self._stop_when_triggered(self._stop_event),
                )

                # Monitor for the queue being empty
                monitor_empty = asyncio.create_task(self._monitor_queue_empty())
                if end_on_empty:
                    self.on_empty(lambda: monitor_empty.cancel(), hidden=True)

                # The timeout criterion is satisfied by the `timeout` arg
                await asyncio.wait(
                    [stop_triggered, monitor_empty],
                    timeout=timeout,
                    return_when=asyncio.FIRST_COMPLETED,
                )

                # Determine the reason for stopping
                stop_reason: BaseException | ExitState.Code
                if stop_triggered.done() and self._stop_event.is_set():
                    stop_reason = ExitState.Code.STOPPED

                    msg, exception = self._stop_event.context
                    _log = logger.exception if exception else logger.debug
                    _log(f"Stop Message: {msg}", exc_info=exception)

                    self.on_stop.emit(str(msg), exception)
                    if self._on_exc_method_map and exception:
                        stop_reason = exception
                    else:
                        stop_reason = ExitState.Code.STOPPED
                elif monitor_empty.done():
                    logger.debug("Scheduler stopped due to being empty.")
                    stop_reason = ExitState.Code.EXHAUSTED
                elif timeout is not None:
                    logger.debug(f"Scheduler stopping as {timeout=} reached.")
                    stop_reason = ExitState.Code.TIMEOUT
                    self.on_timeout.emit()
                else:
                    logger.warning("Scheduler stopping for unknown reason!")
                    stop_reason = ExitState.Code.UNKNOWN

                # Stop all running async tasks, i.e. monitoring the queue to trigger
                # an event
                tasks = [monitor_empty, stop_triggered]
                for task in tasks:
                    task.cancel()

                # Await all the cancelled tasks and read the exceptions
                await asyncio.gather(*tasks, return_exceptions=True)

                self.on_finishing.emit()
                logger.debug("Scheduler is finished")
                logger.debug(f"Shutting down scheduler executor with {wait=}")

            # Scheduler is now refusing jobs
            logger.debug("Scheduler has shutdown and declared as no longer running")

            # This will try to end the tasks based on wait and self._terminate
            if not wait:
                if self._terminate is None:
                    logger.warning(
                        "Cancelling currently running tasks and then waiting "
                        f" as there is no termination strategy for {self.executor=}`.",
                    )

                for future in self.queue:
                    if not future.done():
                        logger.debug(f"Cancelling {future=}")
                        future.cancel()

                if self._terminate is not None:
                    logger.debug(f"Terminating workers with {termination_strategy=}")
                    self._terminate(self.executor)

            logger.debug("Waiting for executor to finish.")
            self.executor.shutdown(wait=wait)
            self.on_finished.emit()
            logger.debug(f"Scheduler finished with status {stop_reason}")

            # Clear all events
            self._stop_event.clear()
            self._queue_has_items_event.clear()

            if self._live_output is not None:
                self._live_output.refresh()

            if self._timeout_timer is not None:
                self._timeout_timer.cancel()

        return stop_reason

    def run(
        self,
        *,
        timeout: float | None = None,
        end_on_empty: bool = True,
        wait: bool = True,
        on_exception: (
            Literal["raise", "end", "continue"]
            | Mapping[type[Exception], Literal["raise", "end", "continue"]]
        ) = "raise",
        on_cancelled: Literal["raise", "end", "continue"] = "raise",
        asyncio_debug_mode: bool = False,
        display: bool | Iterable[RenderableType] | Literal["auto"] = "auto",
    ) -> ExitState:
        """Run the scheduler.

        Args:
            timeout: The maximum time to run the scheduler for in
                seconds. Defaults to `None` which means no timeout and it
                will end once the queue is empty if `end_on_empty=True`.
            end_on_empty: Whether to end the scheduler when the queue becomes empty.
            wait: Whether to wait for currently running compute to finish once
                the `Scheduler` is shutting down.

                * If `#!python True`, will wait for all currently running compute.
                * If `#!python False`, will attempt to cancel/terminate all currently
                    running compute and shutdown the executor. This may be useful
                    if you want to end the scheduler as quickly as possible or
                    respect the `timeout=` more precisely.
            on_exception: What to do when an exception occurs in the scheduler
                or callbacks (**Does not apply to submitted compute!**)

                * If `#!python "raise"`, the scheduler will stop and raise the
                    exception at the point where you called `run()`.
                * If `#!python "continue"`, the scheduler will continue to run
                    and just emit the exception as usual.
                    This may be useful when requiring more robust execution.
                * If `#!python "end"`, similar to `#!python "raise"`, the scheduler
                    will stop but no exception will occur and the control flow
                    will return gracefully to the point where you called `run()`.

                * If a `dict`, then it's a mapping from error types to what should be
                    done about them. This can be used with something like the
                    [`amltk.scheduling.plugins.plugins.PynisherPlugin`] to prevent
                    it from killing the scheduler when an individual trial fails.

                    ```python
                    from amltk.scheduling.plugins import PynisherPlugin
                    from amltk.scheduling import Scheduler

                    scheduler = Scheduler.with_processes()

                    scheduler.run(
                        on_exception={
                            PynisherPlugin.Exception: "continue",
                            Exception: "raise"  # Default...
                        }
                    )
                    ```

                    !!! tip "Order"

                        The order of the mapping matters. The first matching exception
                        with an `isinstance` check will be used. You should place
                        more explicit exception types first.

                    !!! warning "KeyboardInterrupt"

                        The `KeyboardInterrupt` or `BaseException` can not be
                        caught and will always be raised, regardless of the
                        `on_exception` setting.

            on_cancelled:
                What to do when a task is cancelled.

                * If `#!python "raise"`, the scheduler will stop and raise the
                    exception at the point where you called `run()`.
                * If `#!python "continue"`, the scheduler will continue running,
                    ignoring the cancellation. This may be useful when requiring more
                    robust execution.
                * If `#!python "end"`, similar to `#!python "raise"`, the scheduler
                    will stop but no exception will occur and the control flow
                    will return gracefully to the point where you called `run()`.

            asyncio_debug_mode: Whether to run the async loop in debug mode.
                Defaults to `False`. Please see [asyncio.run][] for more.
            display: Whether to display the scheduler live in the console.

                * If `#!python "auto"`, will display the scheduler if in
                    a notebook or colab environemnt. Otherwise, it will not display
                    it. If left as "auto" and the display occurs, a warning will
                    be printed alongside it.
                * If `#!python False`, will not display anything.
                * If `#!python True`, will display the scheduler and all its tasks.
                * If a `#!python list[RenderableType]` , will display the scheduler
                    itself plus those renderables.

        Returns:
            The reason for the scheduler ending.

        Raises:
            RuntimeError: If the scheduler is already running.
        """
        if display == "auto":
            from amltk._richutil import is_jupyter

            display = is_jupyter()
            if display is True:
                warnings.warn(
                    "Detected that current running context is in a notebook!"
                    " When `display='auto'`, the default, the scheduler will"
                    " automatically be set to display. If you do not want this or"
                    " wish to disable this warning, please set `display=False`.",
                    UserWarning,
                    stacklevel=2,
                )

        return asyncio.run(
            self.async_run(
                timeout=timeout,
                end_on_empty=end_on_empty,
                wait=wait,
                on_exception=on_exception,
                display=display,
                on_cancelled=on_cancelled,
            ),
            debug=asyncio_debug_mode,
        )

    async def async_run(  # noqa: C901, PLR0915
        self,
        *,
        timeout: float | None = None,
        end_on_empty: bool = True,
        wait: bool = True,
        on_exception: (
            Literal["raise", "end", "continue"]
            | Mapping[type[Exception], Literal["raise", "end", "continue"]]
        ) = "raise",
        on_cancelled: Literal["raise", "end", "continue"] = "raise",
        display: bool | Iterable[RenderableType] = False,
    ) -> ExitState:
        """Async version of `run`.

        This can be useful if you are already running in an async context,
        such as in a web server or Jupyter notebook.

        Please see [`run()`][amltk.Scheduler.run] for more details.
        """
        if self.running():
            raise RuntimeError("Scheduler already seems to be running")

        match on_cancelled:
            case "raise" | "end" | "continue":
                self._on_cancel_method.set(value=on_cancelled)
            case _:
                raise ValueError(
                    f"Invalid value for `on_cancelled`: {on_cancelled}."
                    " Please provide one of `raise`, `end`, or `ignore`",
                )

        match on_exception:
            case "raise" | "end" | "continue":
                self._on_exc_method_map.set(value={Exception: on_exception})
            case Mapping():
                self._on_exc_method_map.set(value=dict(on_exception))
                for k, v in on_exception.items():
                    if k in (BaseException, KeyboardInterrupt):
                        raise ValueError(
                            f"Invalid key in `on_exception` mapping: {k}. "
                            "Must not be `BaseException` or `KeyboardInterrupt`.",
                        )

                    if not issubclass(k, Exception):
                        raise ValueError(
                            f"Invalid key in `on_exception` mapping: {k}. "
                            "Must be a subclass of `Exception`.",
                        )

                    if v not in ("raise", "end", "continue"):
                        raise ValueError(
                            f"Invalid value in `on_exception` mapping: {v}. "
                            "Must be one of `raise`, `end`, or `ignore`.",
                        )
            case _:
                raise ValueError(
                    f"Invalid value for `on_exception`: {on_exception}."
                    " Please provide one of `raise`, `end`, or `ignore`"
                    " or a `Mapping` from exception types to one of those values.",
                )

        # If the user has requested to have a live display,
        # we will need to setup a `Live` instance to render to
        match display:
            case False:
                self._extra_renderables = None
                self._live_output = None
            case True:
                from rich.live import Live

                self._extra_renderables = None
                self._live_output = Live(
                    auto_refresh=False,
                    get_renderable=self.__rich__,
                )
            case Iterable():
                from rich.live import Live

                self._extra_renderables = list(display)
                self._live_output = Live(
                    auto_refresh=False,
                    get_renderable=self.__rich__,
                )
            case _:
                raise ValueError(
                    f"Invalid value for `display`: {display}."
                    " Please provide one of `True`, `False`, or an iterable of"
                    " `RenderableType`.",
                )

        loop = asyncio.get_running_loop()

        # If it's been specified that we might need to abruptly end
        # the asyncio handler due to an error (as indicated by "raise" or "end")
        # then we need to setup a custom exception handler that tells the scheduler
        # to stop.
        # By default, the asyncio handler will silently eat exceptions which
        # is the desired behaviour if everything is just "continue"
        previous_exception_handler = None
        if on_exception in ("raise", "end") or (
            isinstance(on_exception, Mapping)
            and any(v in ("raise", "end") for v in on_exception.values())
        ):
            previous_exception_handler = loop.get_exception_handler()

            def custom_exception_handler(
                loop: asyncio.AbstractEventLoop,
                context: dict[str, Any],
            ) -> None:
                # The exception could be None as specified by:
                # https://docs.python.org/3/library/asyncio-eventloop.html#asyncio.loop.call_exception_handler
                # Not sure when exactly this happens but since there's no exception
                # we should just ignore it.
                exception: BaseException | None = context.get("exception")
                message: str | None = context.get("message")

                # handle with previous handler before possibly
                # stopping
                if previous_exception_handler:
                    previous_exception_handler(loop, context)
                else:
                    loop.default_exception_handler(context)

                match exception:
                    case None:
                        return
                    # We do not allow special handling of BaseException by users
                    case BaseException() if not isinstance(exception, Exception):
                        self.stop(stop_msg=message, exception=exception)
                        return
                    case Exception():
                        match subclass_map(exception, self._on_exc_method_map.value):
                            case None | (_, "continue"):
                                return
                            case (_, "raise") | (_, "end"):
                                self.stop(stop_msg=message, exception=exception)
                                return
                            case _:
                                raise RuntimeError("Unexpected, please raise issue!")
                    case _:
                        msg = (
                            "Recieved something odd."
                            " Please raise an issue with this info!"
                            f"{type(exception)} {exception=}"
                            f"\n{context=}"
                            f"\n{loop=}"
                            "\n===\n"
                            f"{message}"
                        )
                        self.stop(stop_msg=msg, exception=exception)
                        return

            loop.set_exception_handler(custom_exception_handler)

        logger.debug("Starting scheduler")
        # Run the actual scheduling loop
        stop_reason = await self._run_scheduler(
            timeout=timeout,
            end_on_empty=end_on_empty,
            wait=wait,
        )
        if previous_exception_handler is not None:
            loop.set_exception_handler(previous_exception_handler)

        # Need a reference to later this before resetting...
        exc_method_map = dict(self._on_exc_method_map.value)

        # Reset variables back to its default
        self._live_output = None
        self._extra_renderables = None
        self._on_exc_method_map.reset()
        self._on_cancel_method.reset()

        # Decide what to do with the result from the scheduler
        match stop_reason:
            case ExitState.Code():
                return ExitState(code=stop_reason)
            # Special case for handling the cancellation error as it does
            # not occur as an exception as far as futures are concerened
            case CancelledError() if on_cancelled == "raise":
                raise stop_reason
            case CancelledError() if on_cancelled in ("end", "continue"):
                return ExitState(code=ExitState.Code.CANCELLED, exception=stop_reason)
            # General exception handling
            case Exception():
                match subclass_map(stop_reason, exc_method_map):
                    case None | (_, "raise"):
                        raise stop_reason
                    case (_, "end"):
                        return ExitState(
                            ExitState.Code.EXCEPTION,
                            exception=stop_reason,
                        )
                    case (_, "continue"):
                        logger.warning(
                            f"Was told to ignore `{type(stop_reason)}` yet it"
                            " was returned by the Scheduler. Please raise an issue!"
                            "\nReturning ExitState.Code.UNKNOWN but this behaviour",
                            " may crash in the future.",
                        )
                        return ExitState(
                            ExitState.Code.UNKNOWN,
                            exception=stop_reason,
                        )
                    case _:
                        raise RuntimeError("Unexpected, please raise issue!")
            case BaseException():
                raise stop_reason

    def stop(
        self,
        *args: Any,
        stop_msg: str | None = None,
        exception: BaseException | None = None,
        **kwargs: Any,
    ) -> None:
        """Stop the scheduler.

        The scheduler will stop, finishing currently running tasks depending
        on the `wait=` parameter to [`Scheduler.run`][amltk.Scheduler.run].

        The call signature is kept open with `*args, **kwargs` to make it
        easier to include in any callback.

        Args:
            *args: Logged in a debug message
            **kwargs: Logged in a debug message
            stop_msg: The message to log when stopping the scheduler.
            exception: The exception which incited `stop()` to be called.
                Will be used by the `Scheduler` to possibly raise the exception
                to the user.
        """
        if not self.running():
            return

        assert self._stop_event is not None

        msg = stop_msg if stop_msg is not None else "scheduler.stop() was called."
        logger.debug(f"Stopping scheduler: {msg} {args=} {kwargs=}")

        self._stop_event.set(msg=msg, exception=exception)
        self._running_event.clear()

    def call_later(
        self,
        delay: float,
        fn: Callable[P, Any],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> asyncio.TimerHandle:
        """Schedule a function to be run after a delay.

        Args:
            delay: The delay in seconds.
            fn: The function to run.
            args: The positional arguments to pass to the function.
            kwargs: The keyword arguments to pass to the function.

        Returns:
            A timer handle that can be used to cancel the function.
        """
        if not self.running():
            raise RuntimeError("Scheduler is not running!")

        _fn = partial(fn, *args, **kwargs)
        loop = asyncio.get_running_loop()
        return loop.call_later(delay, _fn)

    def add_renderable(self, renderable: RenderableType) -> None:
        """Add a renderable object to the scheduler.

        This will be displayed whenever the scheduler is displayed.
        """
        self._renderables.append(renderable)

    @override
    def __rich__(self) -> RenderableType:
        from rich.console import Group
        from rich.panel import Panel
        from rich.pretty import Pretty
        from rich.table import Column, Table
        from rich.text import Text
        from rich.tree import Tree

        from amltk._richutil import richify
        from amltk._richutil.renderers.function import Function

        MAX_FUTURE_ITEMS = 5
        OFFSETS = 1 + 1 + 2  # Header + ellipses space + panel borders

        title = Text("Scheduler", style="magenta bold")
        if self.running():
            title.append(" (running)", style="green")

        future_table = Table.grid()

        # Select the most latest items
        future_items = list(self.queue.items())[-MAX_FUTURE_ITEMS:]
        for future, (func, args, kwargs) in future_items:
            entry = Function(
                func,
                (args, kwargs),
                link=False,
                prefix=future._state,
                no_wrap=True,
            )
            future_table.add_row(entry)

        if len(self.queue) > MAX_FUTURE_ITEMS:
            future_table.add_row(Text("...", style="yellow"))

        queue_column_text = Text.assemble(
            "Queue: (",
            (f"{len(self.queue)}", "yellow"),
            ")",
        )

        layout_table = Table(
            Column("Executor", ratio=1),
            Column(queue_column_text, ratio=2),
            box=None,
            expand=True,
            padding=(0, 1),
        )
        layout_table.add_row(richify(self.executor, otherwise=Pretty), future_table)

        panel = Panel(
            layout_table,
            title=title,
            title_align="left",
            border_style="magenta",
            height=MAX_FUTURE_ITEMS + OFFSETS,
        )
        tree = Tree(panel, guide_style="magenta bold")

        for renderable in self._renderables:
            tree.add(renderable)

        if not self._extra_renderables:
            return tree

        return Group(tree, *self._extra_renderables)


@dataclass
class ExitState:
    """The exit state of a scheduler.

    Attributes:
        reason: The reason for the exit.
        exception: The exception that caused the exit, if any.
    """

    code: ExitState.Code
    exception: BaseException | None = None

    class Code(Enum):
        """The reason the scheduler ended."""

        STOPPED = auto()
        """The scheduler was stopped forcefully with `Scheduler.stop`."""

        TIMEOUT = auto()
        """The scheduler finished because of a timeout."""

        EXHAUSTED = auto()
        """The scheduler finished because it exhausted its queue."""

        CANCELLED = auto()
        """The scheduler was cancelled."""

        UNKNOWN = auto()
        """The scheduler finished for an unknown reason."""

        EXCEPTION = auto()
        """The scheduler finished because of an exception."""



"""Data utilities for scikit-learn."""
from __future__ import annotations

from collections.abc import Sequence
from itertools import chain
from typing import TYPE_CHECKING

from more_itertools import last
from sklearn.model_selection import train_test_split

from amltk.randomness import as_int

if TYPE_CHECKING:
    from amltk.types import Seed


def split_data(
    *items: Sequence,
    splits: dict[str, float],
    seed: Seed | None = None,
    shuffle: bool = True,
    stratify: Sequence | None = None,
) -> dict[str, tuple[Sequence, ...]]:
    """Split a set of items into multiple splits.

    ```python
    from amltk.sklearn.data import split

    x = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
    y = [0, 0, 0, 0, 1, 1, 1, 1, 1, 1]
    splits = split_data(x, y, splits={"train": 0.6, "val": 0.2, "test": 0.2})

    train_x, train_y = splits["train"]
    val_x, val_y = splits["val"]
    test_x, test_y = splits["test"]
    ```

    Args:
        items: The items to split. Must be indexible, like a list, np.ndarray,
            pandas dataframe/series or a tuple, etc...
        splits: A dictionary of split names and their percentage of the data.
            The percentages must sum to 1.
        seed: The seed to use for the random state.
        shuffle: Whether to shuffle the data before splitting. Passed forward
            to [sklearn.model_selection.train_test_split][].
        stratify: The stratification to use for the split. This will be passed
            forward to [sklearn.model_selection.train_test_split][]. We account
            for using the stratification for all splits, ensuring we split of
            the stratification values themselves.

    Returns:
        A dictionary of split names and their split items.
    """
    if not all(0 < s < 1 for s in splits.values()):
        raise ValueError(f"Splits ({splits=}) must be between 0 and 1")

    if sum(splits.values()) != 1:
        raise ValueError(f"Splits ({splits=}) must sum to 1")

    if len(splits) < 2:  # noqa: PLR2004
        raise ValueError(f"Splits ({splits=}) must have at least 2 splits")

    rng = as_int(seed) if seed is not None else None

    # Store the results of each split, indexed by the split number
    split_results: dict[str, list[Sequence]] = {}
    remaining: list[Sequence] = list(items)

    remaining_percentage = 1.0

    # Enumerate up to the last split
    for name, split_percentage in list(splits.items())[0:-1]:
        # If we stratify, make sure to also include it in the splitting so
        # further splits can be stratified correctly.
        to_split = remaining if stratify is None else [*remaining, stratify]

        # Calculate the percentage of the remaining data to split
        percentage = split_percentage / remaining_percentage

        splitted = train_test_split(
            *to_split,
            train_size=percentage,
            random_state=rng,
            shuffle=shuffle,
            stratify=stratify,
        )

        # Update the remaining percentage
        remaining_percentage -= split_percentage

        # Splitted returns pairs of (train, test) for each item in items
        # so we need to split them up
        lefts = splitted[::2]
        rights = splitted[1::2]

        # If we had stratify, we need to remove the last item from splits
        # as it was the stratified array, setting the stratification for
        # the next split
        if stratify is not None:
            stratify = rights[-1]  # type: ignore

            lefts = lefts[:-1]
            rights = rights[:-1]

        # Lastly, we insert the lefts into the split_results
        # and set the remaining to the rights
        split_results[name] = lefts  # type: ignore
        remaining = rights  # type: ignore

    # Since we enumerated up to the last split, we need to add the last
    # split manually
    last_name = last(splits.keys())
    split_results[last_name] = remaining

    return {name: tuple(split) for name, split in split_results.items()}


def train_val_test_split(
    *items: Sequence,
    splits: tuple[float, float, float],
    seed: Seed | None = None,
    shuffle: bool = True,
    stratify: Sequence | None = None,
) -> tuple[Sequence, ...]:
    """Split a set of items into train, val and test splits.

    ```python
    from amltk.sklearn.data import train_val_test_split

    x = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
    y = [0, 0, 0, 0, 1, 1, 1, 1, 1, 1]
    train_x, train_y, val_x, val_y, test_x, test_y = train_val_test_split(
        x, y, splits=(0.6, 0.2, 0.2),
    )
    ```

    Args:
        items: The items to split. Must be indexible, like a list, np.ndarray,
            pandas dataframe/series or a tuple, etc...
        splits: A tuple of the percentage of the data to use for the train,
            val and test splits. The percentages must sum to 1.
        seed: The seed to use for the random state.
        shuffle: Whether to shuffle the data before splitting. Passed forward
            to [sklearn.model_selection.train_test_split][].
        stratify: The stratification to use for the split. This will be passed
            forward to [sklearn.model_selection.train_test_split][]. We account
            for using the stratification for all splits, ensuring we split of
            the stratification values themselves.

    Returns:
        A tuple containing the train, val and test splits.
    """
    results = split_data(
        *items,
        splits={"train": splits[0], "val": splits[1], "test": splits[2]},
        seed=seed,
        shuffle=shuffle,
        stratify=stratify,
    )
    return tuple(chain(results["train"], results["val"], results["test"]))
